"""Module for intercepting stdout/stderr.

This patches the `write()` method of `stdout` and `stderr` on import.
Once patched, it is not possible to unpatch or repatch, though individual
callbacks can be removed.

We assume that all other writing methods on the object delegate to `write()`,
like `writelines()`. This is not guaranteed to be true, but it is true for
common implementations. In particular, CPython's implementation of IOBase's
`writelines()` delegates to `write()`.

It is important to note that this technique interacts poorly with other
code that performs similar patching if it also allows unpatching as this
discards our modification. This is why we patch on import and do not support
unpatching:

    with contextlib.redirect_stderr(...):
        from ... import console_capture
        # Here, everything works fine.
    # Here, callbacks are never called again.

In particular, it does not work with some combinations of pytest's
`capfd` / `capsys` fixtures and pytest's `--capture` option.
"""

from __future__ import annotations

import sys
import threading
from typing import IO, AnyStr, Callable, Protocol


class CannotCaptureConsoleError(Exception):
    """The module failed to patch stdout or stderr."""


class _WriteCallback(Protocol):
    """A callback that receives intercepted bytes or string data."""

    def __call__(
        self,
        data: bytes | str,
        written: int,
        /,
    ) -> None:
        """Intercept data passed to `write()`.

        Args:
            data: The object passed to stderr's or stdout's `write()`.
            written: The number of bytes or characters written.
                This is the return value of `write()`.
        """


_module_lock = threading.Lock()

_patch_exception: CannotCaptureConsoleError | None = None

_next_callback_id: int = 1

_stdout_callbacks: dict[int, _WriteCallback] = {}
_stderr_callbacks: dict[int, _WriteCallback] = {}


def capture_stdout(callback: _WriteCallback) -> Callable[[], None]:
    """Install a callback that runs after every write to sys.stdout.

    Args:
        callback: A callback to invoke after running `sys.stdout.write`.
            This may be called from any thread, so it must be thread-safe.
            Exceptions are propagated to the caller of `write`.
            See `_WriteCallback` for the exact protocol.

    Returns:
        A function to uninstall the callback.

    Raises:
        CannotCaptureConsoleError: If patching failed on import.
    """
    with _module_lock:
        if _patch_exception:
            raise _patch_exception

        return _insert_disposably(
            _stdout_callbacks,
            callback,
        )


def capture_stderr(callback: _WriteCallback) -> Callable[[], None]:
    """Install a callback that runs after every write to sys.sdterr.

    Args:
        callback: A callback to invoke after running `sys.stderr.write`.
            This may be called from any thread, so it must be thread-safe.
            Exceptions are propagated to the caller of `write`.
            See `_WriteCallback` for the exact protocol.

    Returns:
        A function to uninstall the callback.

    Raises:
        CannotCaptureConsoleError: If patching failed on import.
    """
    with _module_lock:
        if _patch_exception:
            raise _patch_exception

        return _insert_disposably(
            _stderr_callbacks,
            callback,
        )


def _insert_disposably(
    callback_dict: dict[int, _WriteCallback],
    callback: _WriteCallback,
) -> Callable[[], None]:
    global _next_callback_id
    id = _next_callback_id
    _next_callback_id += 1

    disposed = False

    def dispose() -> None:
        nonlocal disposed

        with _module_lock:
            if disposed:
                return

            del callback_dict[id]

            disposed = True

    callback_dict[id] = callback
    return dispose


def _patch(
    stdout_or_stderr: IO[AnyStr],
    callbacks: dict[int, _WriteCallback],
) -> None:
    orig_write: Callable[[AnyStr], int]

    def write_with_callbacks(s: AnyStr, /) -> int:
        n = orig_write(s)

        # We make a copy here because callbacks could, in theory, modify
        # the list of callbacks.
        with _module_lock:
            callbacks_copy = list(callbacks.values())

        for cb in callbacks_copy:
            cb(s, n)

        return n

    orig_write = stdout_or_stderr.write

    # mypy==1.14.1 fails to type-check this:
    #   Incompatible types in assignment (expression has type
    #   "Callable[[bytes], int]", variable has type overloaded function)
    stdout_or_stderr.write = write_with_callbacks  # type: ignore


try:
    _patch(sys.stdout, _stdout_callbacks)
    _patch(sys.stderr, _stderr_callbacks)
except Exception as _patch_exception_cause:
    _patch_exception = CannotCaptureConsoleError()
    _patch_exception.__cause__ = _patch_exception_cause



# Generated by ariadne-codegen

from .add_aliases import AddAliases, AddAliasesAddAliases
from .artifact_collection_membership_files import (
    ArtifactCollectionMembershipFiles,
    ArtifactCollectionMembershipFilesProject,
    ArtifactCollectionMembershipFilesProjectArtifactCollection,
    ArtifactCollectionMembershipFilesProjectArtifactCollectionArtifactMembership,
)
from .artifact_version_files import (
    ArtifactVersionFiles,
    ArtifactVersionFilesProject,
    ArtifactVersionFilesProjectArtifactType,
    ArtifactVersionFilesProjectArtifactTypeArtifact,
)
from .create_artifact_collection_tag_assignments import (
    CreateArtifactCollectionTagAssignments,
    CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignments,
    CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignmentsTags,
)
from .delete_aliases import DeleteAliases, DeleteAliasesDeleteAliases
from .delete_artifact_collection_tag_assignments import (
    DeleteArtifactCollectionTagAssignments,
    DeleteArtifactCollectionTagAssignmentsDeleteArtifactCollectionTagAssignments,
)
from .delete_artifact_portfolio import (
    DeleteArtifactPortfolio,
    DeleteArtifactPortfolioDeleteArtifactPortfolio,
    DeleteArtifactPortfolioDeleteArtifactPortfolioArtifactCollection,
)
from .delete_artifact_sequence import (
    DeleteArtifactSequence,
    DeleteArtifactSequenceDeleteArtifactSequence,
    DeleteArtifactSequenceDeleteArtifactSequenceArtifactCollection,
)
from .enums import ArtifactCollectionState, ArtifactState
from .fetch_linked_artifacts import (
    FetchLinkedArtifacts,
    FetchLinkedArtifactsArtifact,
    FetchLinkedArtifactsArtifactArtifactMemberships,
    FetchLinkedArtifactsArtifactArtifactMembershipsEdges,
    FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNode,
    FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNodeAliases,
    FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNodeArtifactCollection,
    FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNodeArtifactCollectionProject,
)
from .fragments import (
    ArtifactCollectionsFragment,
    ArtifactCollectionsFragmentEdges,
    ArtifactCollectionsFragmentEdgesNode,
    ArtifactCollectionsFragmentPageInfo,
    ArtifactFragment,
    ArtifactFragmentAliases,
    ArtifactFragmentAliasesArtifactCollection,
    ArtifactFragmentAliasesArtifactCollectionProject,
    ArtifactFragmentArtifactSequence,
    ArtifactFragmentArtifactSequenceProject,
    ArtifactFragmentArtifactType,
    ArtifactFragmentCurrentManifest,
    ArtifactFragmentCurrentManifestFile,
    ArtifactFragmentTags,
    ArtifactPortfolioTypeFields,
    ArtifactSequenceTypeFields,
    ArtifactsFragment,
    ArtifactsFragmentEdges,
    ArtifactsFragmentPageInfo,
    ArtifactTypeFragment,
    ArtifactTypesFragment,
    ArtifactTypesFragmentEdges,
    ArtifactTypesFragmentPageInfo,
    FilesFragment,
    FilesFragmentEdges,
    FilesFragmentEdgesNode,
    FilesFragmentPageInfo,
)
from .input_types import ArtifactAliasInput, ArtifactCollectionAliasInput, TagInput
from .move_artifact_collection import (
    MoveArtifactCollection,
    MoveArtifactCollectionMoveArtifactSequence,
    MoveArtifactCollectionMoveArtifactSequenceArtifactCollection,
)
from .operations import (
    ADD_ALIASES_GQL,
    ARTIFACT_COLLECTION_MEMBERSHIP_FILES_GQL,
    ARTIFACT_VERSION_FILES_GQL,
    CREATE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL,
    DELETE_ALIASES_GQL,
    DELETE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL,
    DELETE_ARTIFACT_PORTFOLIO_GQL,
    DELETE_ARTIFACT_SEQUENCE_GQL,
    FETCH_LINKED_ARTIFACTS_GQL,
    MOVE_ARTIFACT_COLLECTION_GQL,
    PROJECT_ARTIFACT_COLLECTION_GQL,
    PROJECT_ARTIFACT_COLLECTIONS_GQL,
    PROJECT_ARTIFACT_TYPE_GQL,
    PROJECT_ARTIFACT_TYPES_GQL,
    PROJECT_ARTIFACTS_GQL,
    RUN_INPUT_ARTIFACTS_GQL,
    RUN_OUTPUT_ARTIFACTS_GQL,
    UPDATE_ARTIFACT_GQL,
    UPDATE_ARTIFACT_PORTFOLIO_GQL,
    UPDATE_ARTIFACT_SEQUENCE_GQL,
)
from .project_artifact_collection import (
    ProjectArtifactCollection,
    ProjectArtifactCollectionProject,
    ProjectArtifactCollectionProjectArtifactType,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollection,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliases,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdges,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdgesNode,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesPageInfo,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTags,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdges,
    ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdgesNode,
    ProjectArtifactCollectionProjectArtifactTypeArtifactSequence,
)
from .project_artifact_collections import (
    ProjectArtifactCollections,
    ProjectArtifactCollectionsProject,
    ProjectArtifactCollectionsProjectArtifactType,
)
from .project_artifact_type import ProjectArtifactType, ProjectArtifactTypeProject
from .project_artifact_types import ProjectArtifactTypes, ProjectArtifactTypesProject
from .project_artifacts import (
    ProjectArtifacts,
    ProjectArtifactsProject,
    ProjectArtifactsProjectArtifactType,
    ProjectArtifactsProjectArtifactTypeArtifactCollection,
)
from .run_input_artifacts import (
    RunInputArtifacts,
    RunInputArtifactsProject,
    RunInputArtifactsProjectRun,
    RunInputArtifactsProjectRunInputArtifacts,
    RunInputArtifactsProjectRunInputArtifactsEdges,
    RunInputArtifactsProjectRunInputArtifactsPageInfo,
)
from .run_output_artifacts import (
    RunOutputArtifacts,
    RunOutputArtifactsProject,
    RunOutputArtifactsProjectRun,
    RunOutputArtifactsProjectRunOutputArtifacts,
    RunOutputArtifactsProjectRunOutputArtifactsEdges,
    RunOutputArtifactsProjectRunOutputArtifactsPageInfo,
)
from .update_artifact import UpdateArtifact, UpdateArtifactUpdateArtifact
from .update_artifact_portfolio import (
    UpdateArtifactPortfolio,
    UpdateArtifactPortfolioUpdateArtifactPortfolio,
    UpdateArtifactPortfolioUpdateArtifactPortfolioArtifactCollection,
)
from .update_artifact_sequence import (
    UpdateArtifactSequence,
    UpdateArtifactSequenceUpdateArtifactSequence,
    UpdateArtifactSequenceUpdateArtifactSequenceArtifactCollection,
)

__all__ = [
    "ADD_ALIASES_GQL",
    "ARTIFACT_COLLECTION_MEMBERSHIP_FILES_GQL",
    "ARTIFACT_VERSION_FILES_GQL",
    "CREATE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL",
    "DELETE_ALIASES_GQL",
    "DELETE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL",
    "DELETE_ARTIFACT_PORTFOLIO_GQL",
    "DELETE_ARTIFACT_SEQUENCE_GQL",
    "FETCH_LINKED_ARTIFACTS_GQL",
    "MOVE_ARTIFACT_COLLECTION_GQL",
    "PROJECT_ARTIFACTS_GQL",
    "PROJECT_ARTIFACT_COLLECTIONS_GQL",
    "PROJECT_ARTIFACT_COLLECTION_GQL",
    "PROJECT_ARTIFACT_TYPES_GQL",
    "PROJECT_ARTIFACT_TYPE_GQL",
    "RUN_INPUT_ARTIFACTS_GQL",
    "RUN_OUTPUT_ARTIFACTS_GQL",
    "UPDATE_ARTIFACT_GQL",
    "UPDATE_ARTIFACT_PORTFOLIO_GQL",
    "UPDATE_ARTIFACT_SEQUENCE_GQL",
    "DeleteArtifactSequence",
    "DeleteArtifactSequenceDeleteArtifactSequence",
    "DeleteArtifactSequenceDeleteArtifactSequenceArtifactCollection",
    "DeleteArtifactPortfolio",
    "DeleteArtifactPortfolioDeleteArtifactPortfolio",
    "DeleteArtifactPortfolioDeleteArtifactPortfolioArtifactCollection",
    "UpdateArtifactSequence",
    "UpdateArtifactSequenceUpdateArtifactSequence",
    "UpdateArtifactSequenceUpdateArtifactSequenceArtifactCollection",
    "UpdateArtifactPortfolio",
    "UpdateArtifactPortfolioUpdateArtifactPortfolio",
    "UpdateArtifactPortfolioUpdateArtifactPortfolioArtifactCollection",
    "MoveArtifactCollection",
    "MoveArtifactCollectionMoveArtifactSequence",
    "MoveArtifactCollectionMoveArtifactSequenceArtifactCollection",
    "CreateArtifactCollectionTagAssignments",
    "CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignments",
    "CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignmentsTags",
    "DeleteArtifactCollectionTagAssignments",
    "DeleteArtifactCollectionTagAssignmentsDeleteArtifactCollectionTagAssignments",
    "ProjectArtifactCollections",
    "ProjectArtifactCollectionsProject",
    "ProjectArtifactCollectionsProjectArtifactType",
    "ProjectArtifactCollection",
    "ProjectArtifactCollectionProject",
    "ProjectArtifactCollectionProjectArtifactType",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollection",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliases",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdges",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdgesNode",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesPageInfo",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTags",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdges",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdgesNode",
    "ProjectArtifactCollectionProjectArtifactTypeArtifactSequence",
    "ArtifactVersionFiles",
    "ArtifactVersionFilesProject",
    "ArtifactVersionFilesProjectArtifactType",
    "ArtifactVersionFilesProjectArtifactTypeArtifact",
    "ArtifactCollectionMembershipFiles",
    "ArtifactCollectionMembershipFilesProject",
    "ArtifactCollectionMembershipFilesProjectArtifactCollection",
    "ArtifactCollectionMembershipFilesProjectArtifactCollectionArtifactMembership",
    "ProjectArtifactTypes",
    "ProjectArtifactTypesProject",
    "ProjectArtifactType",
    "ProjectArtifactTypeProject",
    "ProjectArtifacts",
    "ProjectArtifactsProject",
    "ProjectArtifactsProjectArtifactType",
    "ProjectArtifactsProjectArtifactTypeArtifactCollection",
    "RunOutputArtifacts",
    "RunOutputArtifactsProject",
    "RunOutputArtifactsProjectRun",
    "RunOutputArtifactsProjectRunOutputArtifacts",
    "RunOutputArtifactsProjectRunOutputArtifactsEdges",
    "RunOutputArtifactsProjectRunOutputArtifactsPageInfo",
    "RunInputArtifacts",
    "RunInputArtifactsProject",
    "RunInputArtifactsProjectRun",
    "RunInputArtifactsProjectRunInputArtifacts",
    "RunInputArtifactsProjectRunInputArtifactsEdges",
    "RunInputArtifactsProjectRunInputArtifactsPageInfo",
    "FetchLinkedArtifacts",
    "FetchLinkedArtifactsArtifact",
    "FetchLinkedArtifactsArtifactArtifactMemberships",
    "FetchLinkedArtifactsArtifactArtifactMembershipsEdges",
    "FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNode",
    "FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNodeAliases",
    "FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNodeArtifactCollection",
    "FetchLinkedArtifactsArtifactArtifactMembershipsEdgesNodeArtifactCollectionProject",
    "AddAliases",
    "AddAliasesAddAliases",
    "DeleteAliases",
    "DeleteAliasesDeleteAliases",
    "UpdateArtifact",
    "UpdateArtifactUpdateArtifact",
    "ArtifactAliasInput",
    "ArtifactCollectionAliasInput",
    "TagInput",
    "ArtifactCollectionsFragment",
    "ArtifactCollectionsFragmentEdges",
    "ArtifactCollectionsFragmentEdgesNode",
    "ArtifactCollectionsFragmentPageInfo",
    "ArtifactFragment",
    "ArtifactFragmentAliases",
    "ArtifactFragmentAliasesArtifactCollection",
    "ArtifactFragmentAliasesArtifactCollectionProject",
    "ArtifactFragmentArtifactSequence",
    "ArtifactFragmentArtifactSequenceProject",
    "ArtifactFragmentArtifactType",
    "ArtifactFragmentCurrentManifest",
    "ArtifactFragmentCurrentManifestFile",
    "ArtifactFragmentTags",
    "ArtifactPortfolioTypeFields",
    "ArtifactSequenceTypeFields",
    "ArtifactTypeFragment",
    "ArtifactTypesFragment",
    "ArtifactTypesFragmentEdges",
    "ArtifactTypesFragmentPageInfo",
    "ArtifactsFragment",
    "ArtifactsFragmentEdges",
    "ArtifactsFragmentPageInfo",
    "FilesFragment",
    "FilesFragmentEdges",
    "FilesFragmentEdgesNode",
    "FilesFragmentPageInfo",
    "ArtifactCollectionState",
    "ArtifactState",
]



import asyncio
from typing import Callable, TypeVar

from wandb.sdk.lib import asyncio_compat, printer

_T = TypeVar("_T")


def run_async_with_spinner(
    spinner_printer: printer.Printer,
    text: str,
    func: Callable[[], _T],
) -> _T:
    """Run a slow function while displaying a loading icon.

    Args:
        spinner_printer: The printer to use to display text.
        text: The text to display next to the spinner while the function runs.
        func: The function to run.

    Returns:
        The result of func.
    """

    async def _loop_run_with_spinner() -> _T:
        func_running = asyncio.Event()

        async def update_spinner() -> None:
            tick = 0
            with spinner_printer.dynamic_text() as text_area:
                if text_area:
                    while not func_running.is_set():
                        spinner = spinner_printer.loading_symbol(tick)
                        text_area.set_text(f"{spinner} {text}")
                        tick += 1
                        await asyncio.sleep(0.1)
                else:
                    spinner_printer.display(text)

        async with asyncio_compat.open_task_group() as group:
            group.start_soon(update_spinner())
            res = await asyncio.get_running_loop().run_in_executor(None, func)
            func_running.set()
            return res

    return asyncio_compat.run(_loop_run_with_spinner)



"""Router - handle message router (sock).

Router to manage responses from a socket client.

"""

from __future__ import annotations

from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_server_pb2 as spb
from wandb.sdk.lib.sock_client import SockClient, SockClientClosedError
from wandb.sdk.mailbox import Mailbox

from .router import MessageRouter, MessageRouterClosedError


class MessageSockRouter(MessageRouter):
    _sock_client: SockClient
    _mailbox: Mailbox

    def __init__(self, sock_client: SockClient, mailbox: Mailbox) -> None:
        self._sock_client = sock_client
        super().__init__(mailbox=mailbox)

    def _read_message(self) -> spb.ServerResponse | None:
        try:
            return self._sock_client.read_server_response(timeout=1)
        except SockClientClosedError as e:
            raise MessageRouterClosedError from e

    def _send_message(self, record: pb.Record) -> None:
        self._sock_client.send_record_communicate(record)



"""Router - handle message router (base class).

Router to manage responses.

"""

from __future__ import annotations

import logging
import threading
from abc import abstractmethod
from typing import TYPE_CHECKING

from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_server_pb2 as spb
from wandb.sdk import mailbox

if TYPE_CHECKING:
    from queue import Queue


logger = logging.getLogger("wandb")


class MessageRouterClosedError(Exception):
    """Router has been closed."""


class MessageRouter:
    _request_queue: Queue[pb.Record]
    _response_queue: Queue[pb.Result]
    _mailbox: mailbox.Mailbox | None

    def __init__(self, mailbox: mailbox.Mailbox | None = None) -> None:
        self._mailbox = mailbox
        self._lock = threading.Lock()

        self._join_event = threading.Event()
        self._thread = threading.Thread(target=self.message_loop)
        self._thread.name = "MsgRouterThr"
        self._thread.daemon = True
        self._thread.start()

    @abstractmethod
    def _read_message(self) -> pb.Result | spb.ServerResponse | None:
        raise NotImplementedError

    @abstractmethod
    def _send_message(self, record: pb.Record) -> None:
        raise NotImplementedError

    def message_loop(self) -> None:
        try:
            while not self._join_event.is_set():
                try:
                    msg = self._read_message()
                except EOFError:
                    # On abnormal shutdown the queue will be destroyed underneath
                    # resulting in EOFError.  message_loop needs to exit..
                    logger.warning("EOFError seen in message_loop")
                    break
                except MessageRouterClosedError as e:
                    logger.warning("message_loop has been closed", exc_info=e)
                    break
                if not msg:
                    continue
                self._handle_msg_rcv(msg)

        finally:
            if self._mailbox:
                self._mailbox.close()

    def join(self) -> None:
        self._join_event.set()
        self._thread.join()

    def _handle_msg_rcv(self, msg: pb.Result | spb.ServerResponse) -> None:
        if not self._mailbox:
            return

        if isinstance(msg, pb.Result) and msg.control.mailbox_slot:
            self._mailbox.deliver(
                spb.ServerResponse(
                    request_id=msg.control.mailbox_slot,
                    result_communicate=msg,
                )
            )
        elif isinstance(msg, spb.ServerResponse) and msg.request_id:
            self._mailbox.deliver(msg)



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import List, Literal, Optional

from pydantic import Field

from wandb._pydantic import GQLBase, GQLId, Typename


class ProjectArtifactCollection(GQLBase):
    project: Optional[ProjectArtifactCollectionProject]


class ProjectArtifactCollectionProject(GQLBase):
    artifact_type: Optional[ProjectArtifactCollectionProjectArtifactType] = Field(
        alias="artifactType"
    )


class ProjectArtifactCollectionProjectArtifactType(GQLBase):
    artifact_collection: Optional[
        ProjectArtifactCollectionProjectArtifactTypeArtifactCollection
    ] = Field(alias="artifactCollection")
    artifact_sequence: Optional[
        ProjectArtifactCollectionProjectArtifactTypeArtifactSequence
    ] = Field(alias="artifactSequence")


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollection(GQLBase):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    id: GQLId
    name: str
    description: Optional[str]
    created_at: str = Field(alias="createdAt")
    tags: ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTags
    aliases: ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliases


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTags(GQLBase):
    edges: List[ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdges]


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdges(GQLBase):
    node: ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdgesNode


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdgesNode(
    GQLBase
):
    id: GQLId
    name: str


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliases(GQLBase):
    edges: List[
        ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdges
    ]
    page_info: ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesPageInfo = Field(
        alias="pageInfo"
    )


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdges(
    GQLBase
):
    node: Optional[
        ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdgesNode
    ]
    cursor: str


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdgesNode(
    GQLBase
):
    alias: str


class ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesPageInfo(
    GQLBase
):
    end_cursor: Optional[str] = Field(alias="endCursor")
    has_next_page: bool = Field(alias="hasNextPage")


class ProjectArtifactCollectionProjectArtifactTypeArtifactSequence(GQLBase):
    typename__: Typename[Literal["ArtifactSequence"]]


ProjectArtifactCollection.model_rebuild()
ProjectArtifactCollectionProject.model_rebuild()
ProjectArtifactCollectionProjectArtifactType.model_rebuild()
ProjectArtifactCollectionProjectArtifactTypeArtifactCollection.model_rebuild()
ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTags.model_rebuild()
ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionTagsEdges.model_rebuild()
ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliases.model_rebuild()
ProjectArtifactCollectionProjectArtifactTypeArtifactCollectionAliasesEdges.model_rebuild()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import ArtifactCollectionsFragment


class ProjectArtifactCollections(GQLBase):
    project: Optional[ProjectArtifactCollectionsProject]


class ProjectArtifactCollectionsProject(GQLBase):
    artifact_type: Optional[ProjectArtifactCollectionsProjectArtifactType] = Field(
        alias="artifactType"
    )


class ProjectArtifactCollectionsProjectArtifactType(GQLBase):
    artifact_collections: Optional[ArtifactCollectionsFragment] = Field(
        alias="artifactCollections"
    )


ProjectArtifactCollections.model_rebuild()
ProjectArtifactCollectionsProject.model_rebuild()
ProjectArtifactCollectionsProjectArtifactType.model_rebuild()



"""Events that trigger W&B Automations."""

# ruff: noqa: UP007  # Avoid using `X | Y` for union fields, as this can cause issues with pydantic < 2.6

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from pydantic import Field
from typing_extensions import Annotated, Self, get_args

from wandb._pydantic import (
    GQLBase,
    SerializedToJson,
    ensure_json,
    field_validator,
    model_validator,
    pydantic_isinstance,
)

from ._filters import And, MongoLikeFilter, Or
from ._filters.expressions import FilterableField
from ._filters.run_metrics import MetricChangeFilter, MetricThresholdFilter, MetricVal
from ._generated import FilterEventFields
from ._validators import LenientStrEnum, simplify_op
from .actions import InputAction, InputActionTypes, SavedActionTypes
from .scopes import ArtifactCollectionScope, AutomationScope, ProjectScope

if TYPE_CHECKING:
    from .automations import NewAutomation


# NOTE: Re-defined publicly with a more readable name for easier access
class EventType(LenientStrEnum):
    """The type of event that triggers an automation."""

    # ---------------------------------------------------------------------------
    # Events triggered by GraphQL mutations
    UPDATE_ARTIFACT_ALIAS = "UPDATE_ARTIFACT_ALIAS"  # NOTE: Avoid in new automations

    CREATE_ARTIFACT = "CREATE_ARTIFACT"
    ADD_ARTIFACT_ALIAS = "ADD_ARTIFACT_ALIAS"
    LINK_ARTIFACT = "LINK_MODEL"
    # Note: "LINK_MODEL" is the (legacy) value expected by the backend, but we
    # name it "LINK_ARTIFACT" here in the public API for clarity and consistency.

    # ---------------------------------------------------------------------------
    # Events triggered by Run conditions
    RUN_METRIC_THRESHOLD = "RUN_METRIC"
    RUN_METRIC_CHANGE = "RUN_METRIC_CHANGE"


# ------------------------------------------------------------------------------
# Saved types: for parsing response data from saved automations


# Note: In GQL responses containing saved automation data, the filter is wrapped in an extra `filter` key.
class _WrappedSavedEventFilter(GQLBase):  # from: TriggeringFilterEvent
    filter: SerializedToJson[MongoLikeFilter] = And()


class _WrappedMetricFilter(GQLBase):  # from: RunMetricFilter
    threshold_filter: Optional[MetricThresholdFilter] = None
    change_filter: Optional[MetricChangeFilter] = None

    @model_validator(mode="before")
    @classmethod
    def _wrap_metric_filter(cls, v: Any) -> Any:
        if pydantic_isinstance(v, MetricThresholdFilter):
            return cls(threshold_filter=v)
        if pydantic_isinstance(v, MetricChangeFilter):
            return cls(change_filter=v)
        return v

    @model_validator(mode="after")
    def _ensure_exactly_one_set(self) -> Self:
        set_fields = [name for name, val in self if (val is not None)]

        if not set_fields:
            all_names = ", ".join(map(repr, type(self).model_fields))
            raise ValueError(f"Expected one of: {all_names}")

        if len(set_fields) > 1:
            set_names = ", ".join(map(repr, set_fields))
            raise ValueError(f"Expected exactly one metric filter, got: {set_names}")

        return self

    @property
    def event_type(self) -> EventType:
        if self.threshold_filter is not None:
            return EventType.RUN_METRIC_THRESHOLD
        if self.change_filter is not None:
            return EventType.RUN_METRIC_CHANGE
        raise RuntimeError("Expected one of: `threshold_filter` or `change_filter`")


class RunMetricFilter(GQLBase):  # from: TriggeringRunMetricEvent
    run: Annotated[SerializedToJson[MongoLikeFilter], Field(alias="run_filter")] = And()
    metric: Annotated[_WrappedMetricFilter, Field(alias="run_metric_filter")]

    # ------------------------------------------------------------------------------
    legacy_metric_filter: Annotated[
        Optional[SerializedToJson[MetricThresholdFilter]],
        Field(alias="metric_filter", deprecated=True),
    ] = None
    """Deprecated legacy field that was previously used to define run metric threshold events.

    For new automations, use the `metric` field (`run_metric_filter` JSON alias) instead.
    """

    @model_validator(mode="before")
    @classmethod
    def _wrap_metric_filter(cls, v: Any) -> Any:
        if pydantic_isinstance(v, (MetricThresholdFilter, MetricChangeFilter)):
            # If only an (unnested) metric filter is given, nest it under the
            # `metric` field, delegating to inner validator(s) for further
            # wrapping/nesting, if needed.
            # This is necessary to conform to the expected backend schema.
            return cls(metric=v)
        return v

    @field_validator("run", mode="after")
    def _wrap_run_filter(cls, v: MongoLikeFilter) -> Any:
        v_new = simplify_op(v)
        return v_new if pydantic_isinstance(v_new, And) else And(and_=[v_new])


class SavedEvent(FilterEventFields):  # from: FilterEventTriggeringCondition
    """A triggering event from a saved automation."""

    event_type: Annotated[EventType, Field(frozen=True)]  # type: ignore[assignment]

    # We override the type of the `filter` field in order to enforce the expected
    # structure for the JSON data when validating and serializing.
    filter: SerializedToJson[Union[_WrappedSavedEventFilter, RunMetricFilter]]
    """The condition(s) under which this event triggers an automation."""


# ------------------------------------------------------------------------------
# Input types: for creating or updating automations


# Note: The GQL input for "eventFilter" does NOT wrap the filter in an extra `filter` key, unlike the
# eventFilter returned in responses for saved automations.
class _BaseEventInput(GQLBase):
    event_type: EventType

    scope: AutomationScope
    """The scope of the event."""

    filter: SerializedToJson[Any]

    def then(self, action: InputAction) -> NewAutomation:
        """Define a new Automation in which this event triggers the given action."""
        from .automations import NewAutomation

        if isinstance(action, (InputActionTypes, SavedActionTypes)):
            return NewAutomation(event=self, action=action)

        raise TypeError(f"Expected a valid action, got: {type(action).__qualname__!r}")

    def __rshift__(self, other: InputAction) -> NewAutomation:
        """Implements `event >> action` to define an Automation with this event and action."""
        return self.then(other)


# ------------------------------------------------------------------------------
# Events that trigger on specific mutations in the backend
class _BaseMutationEventInput(_BaseEventInput):
    filter: SerializedToJson[MongoLikeFilter] = And()
    """Additional condition(s), if any, that must be met for this event to trigger an automation."""

    @field_validator("filter", mode="after")
    def _wrap_filter(cls, v: Any) -> Any:
        """Ensure the given filter is wrapped like: `{"$or": [{"$and": [<original_filter>]}]}`.

        This is awkward but necessary, because the frontend expects this format.
        """
        v_new = simplify_op(v)
        v_new = v_new if pydantic_isinstance(v_new, And) else And(and_=[v_new])
        return Or(or_=[v_new])


class OnLinkArtifact(_BaseMutationEventInput):
    """A new artifact is linked to a collection."""

    event_type: Literal[EventType.LINK_ARTIFACT] = EventType.LINK_ARTIFACT


class OnAddArtifactAlias(_BaseMutationEventInput):
    """A new alias is assigned to an artifact."""

    event_type: Literal[EventType.ADD_ARTIFACT_ALIAS] = EventType.ADD_ARTIFACT_ALIAS


class OnCreateArtifact(_BaseMutationEventInput):
    """A new artifact is created."""

    event_type: Literal[EventType.CREATE_ARTIFACT] = EventType.CREATE_ARTIFACT

    scope: ArtifactCollectionScope
    """The scope of the event: only artifact collections are valid scopes for this event."""


# ------------------------------------------------------------------------------
# Events that trigger on run conditions
class _BaseRunEventInput(_BaseEventInput):
    scope: ProjectScope
    """The scope of the event: only projects are valid scopes for this event."""


class OnRunMetric(_BaseRunEventInput):
    """A run metric satisfies a user-defined condition."""

    event_type: Literal[EventType.RUN_METRIC_THRESHOLD, EventType.RUN_METRIC_CHANGE]

    filter: SerializedToJson[RunMetricFilter]
    """Run and/or metric condition(s) that must be satisfied for this event to trigger an automation."""

    @model_validator(mode="before")
    @classmethod
    def _infer_event_type(cls, data: Any) -> Any:
        """Infer the event type at validation time from the inner filter.

        This allows this class to accommodate both "threshold" and "change" metric
        filter types, which are can only be determined after parsing and validating
        the inner JSON data.
        """
        if isinstance(data, dict) and (raw_filter := data.get("filter")):
            # At this point, `raw_filter` may or may not be JSON-serialized
            parsed_filter = RunMetricFilter.model_validate_json(ensure_json(raw_filter))
            return {**data, "event_type": parsed_filter.metric.event_type}

        return data


# for type annotations
InputEvent = Annotated[
    Union[
        OnLinkArtifact,
        OnAddArtifactAlias,
        OnCreateArtifact,
        OnRunMetric,
    ],
    Field(discriminator="event_type"),
]
# for runtime type checks
InputEventTypes: tuple[type, ...] = get_args(InputEvent.__origin__)  # type: ignore[attr-defined]


# ----------------------------------------------------------------------------


class RunEvent:
    name = FilterableField(server_name="display_name")
    # `Run.name` is actually filtered on `Run.display_name` in the backend.
    # We can't reasonably expect users to know this a priori, so
    # automatically fix it here.

    @staticmethod
    def metric(name: str) -> MetricVal:
        """Define a metric filter condition."""
        return MetricVal(name=name)


class ArtifactEvent:
    alias = FilterableField()


MetricThresholdFilter.model_rebuild()
RunMetricFilter.model_rebuild()
_WrappedSavedEventFilter.model_rebuild()

OnLinkArtifact.model_rebuild()
OnAddArtifactAlias.model_rebuild()
OnCreateArtifact.model_rebuild()
OnRunMetric.model_rebuild()

__all__ = [
    "EventType",
    *(cls.__name__ for cls in InputEventTypes),
    "RunEvent",
    "ArtifactEvent",
    "MetricThresholdFilter",
    "MetricChangeFilter",
]



"""Public API: artifacts."""

import json
import re
from copy import copy
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
)

from typing_extensions import override
from wandb_gql import Client, gql

import wandb
from wandb.apis import public
from wandb.apis.normalize import normalize_exceptions
from wandb.apis.paginator import Paginator, SizedPaginator
from wandb.errors.term import termlog
from wandb.proto.wandb_deprecated import Deprecated
from wandb.proto.wandb_internal_pb2 import ServerFeature
from wandb.sdk.artifacts._generated import (
    ARTIFACT_COLLECTION_MEMBERSHIP_FILES_GQL,
    ARTIFACT_VERSION_FILES_GQL,
    CREATE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL,
    DELETE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL,
    DELETE_ARTIFACT_PORTFOLIO_GQL,
    DELETE_ARTIFACT_SEQUENCE_GQL,
    MOVE_ARTIFACT_COLLECTION_GQL,
    PROJECT_ARTIFACT_COLLECTION_GQL,
    PROJECT_ARTIFACT_COLLECTIONS_GQL,
    PROJECT_ARTIFACT_TYPE_GQL,
    PROJECT_ARTIFACT_TYPES_GQL,
    PROJECT_ARTIFACTS_GQL,
    RUN_INPUT_ARTIFACTS_GQL,
    RUN_OUTPUT_ARTIFACTS_GQL,
    UPDATE_ARTIFACT_PORTFOLIO_GQL,
    UPDATE_ARTIFACT_SEQUENCE_GQL,
    ArtifactCollectionMembershipFiles,
    ArtifactCollectionsFragment,
    ArtifactsFragment,
    ArtifactTypeFragment,
    ArtifactTypesFragment,
    ArtifactVersionFiles,
    FilesFragment,
    ProjectArtifactCollection,
    ProjectArtifactCollections,
    ProjectArtifacts,
    ProjectArtifactType,
    ProjectArtifactTypes,
    RunInputArtifactsProjectRunInputArtifacts,
    RunOutputArtifactsProjectRunOutputArtifacts,
)
from wandb.sdk.artifacts._graphql_fragments import omit_artifact_fields
from wandb.sdk.artifacts._validators import (
    SOURCE_ARTIFACT_COLLECTION_TYPE,
    validate_artifact_name,
)
from wandb.sdk.internal.internal_api import Api as InternalApi
from wandb.sdk.lib import deprecate

from .utils import gql_compat

if TYPE_CHECKING:
    from wandb.apis.public import RetryingClient, Run


class ArtifactTypes(Paginator["ArtifactType"]):
    QUERY = gql(PROJECT_ARTIFACT_TYPES_GQL)

    last_response: Optional[ArtifactTypesFragment]

    def __init__(
        self,
        client: Client,
        entity: str,
        project: str,
        per_page: int = 50,
    ):
        self.entity = entity
        self.project = project

        variable_values = {
            "entityName": entity,
            "projectName": project,
        }
        super().__init__(client, variable_values, per_page)

    @override
    def _update_response(self) -> None:
        """Fetch and validate the response data for the current page."""
        data = self.client.execute(self.QUERY, variable_values=self.variables)
        result = ProjectArtifactTypes.model_validate(data)

        # Extract the inner `*Connection` result for faster/easier access.
        if not ((proj := result.project) and (conn := proj.artifact_types)):
            raise ValueError(f"Unable to parse {type(self).__name__!r} response data")

        self.last_response = ArtifactTypesFragment.model_validate(conn)

    @property
    def length(self) -> None:
        # TODO
        return None

    @property
    def more(self) -> bool:
        if self.last_response is None:
            return True
        return self.last_response.page_info.has_next_page

    @property
    def cursor(self) -> Optional[str]:
        if self.last_response is None:
            return None
        return self.last_response.edges[-1].cursor

    def update_variables(self) -> None:
        self.variables.update({"cursor": self.cursor})

    def convert_objects(self) -> List["ArtifactType"]:
        if self.last_response is None:
            return []

        return [
            ArtifactType(
                client=self.client,
                entity=self.entity,
                project=self.project,
                type_name=node.name,
                attrs=node.model_dump(exclude_unset=True),
            )
            for edge in self.last_response.edges
            if edge.node and (node := ArtifactTypeFragment.model_validate(edge.node))
        ]


class ArtifactType:
    def __init__(
        self,
        client: Client,
        entity: str,
        project: str,
        type_name: str,
        attrs: Optional[Mapping[str, Any]] = None,
    ):
        self.client = client
        self.entity = entity
        self.project = project
        self.type = type_name
        self._attrs = attrs
        if self._attrs is None:
            self.load()

    def load(self):
        data: Optional[Mapping[str, Any]] = self.client.execute(
            gql(PROJECT_ARTIFACT_TYPE_GQL),
            variable_values={
                "entityName": self.entity,
                "projectName": self.project,
                "artifactTypeName": self.type,
            },
        )
        result = ProjectArtifactType.model_validate(data)
        if not ((proj := result.project) and (artifact_type := proj.artifact_type)):
            raise ValueError(f"Could not find artifact type {self.type}")

        self._attrs = artifact_type.model_dump(exclude_unset=True)
        return self._attrs

    @property
    def id(self):
        return self._attrs["id"]

    @property
    def name(self):
        return self._attrs["name"]

    @normalize_exceptions
    def collections(self, per_page=50):
        """Artifact collections."""
        return ArtifactCollections(self.client, self.entity, self.project, self.type)

    def collection(self, name):
        return ArtifactCollection(
            self.client, self.entity, self.project, name, self.type
        )

    def __repr__(self):
        return f"<ArtifactType {self.type}>"


class ArtifactCollections(SizedPaginator["ArtifactCollection"]):
    last_response: Optional[ArtifactCollectionsFragment]

    def __init__(
        self,
        client: Client,
        entity: str,
        project: str,
        type_name: str,
        per_page: int = 50,
    ):
        self.entity = entity
        self.project = project
        self.type_name = type_name

        variable_values = {
            "entityName": entity,
            "projectName": project,
            "artifactTypeName": type_name,
        }

        if server_supports_artifact_collections_gql_edges(client):
            rename_fields = None
        else:
            rename_fields = {"artifactCollections": "artifactSequences"}

        self.QUERY = gql_compat(
            PROJECT_ARTIFACT_COLLECTIONS_GQL, rename_fields=rename_fields
        )

        super().__init__(client, variable_values, per_page)

    @override
    def _update_response(self) -> None:
        """Fetch and validate the response data for the current page."""
        data = self.client.execute(self.QUERY, variable_values=self.variables)
        result = ProjectArtifactCollections.model_validate(data)

        # Extract the inner `*Connection` result for faster/easier access.
        if not (
            (proj := result.project)
            and (type_ := proj.artifact_type)
            and (conn := type_.artifact_collections)
        ):
            raise ValueError(f"Unable to parse {type(self).__name__!r} response data")

        self.last_response = ArtifactCollectionsFragment.model_validate(conn)

    @property
    def length(self):
        if self.last_response is None:
            return None
        return self.last_response.total_count

    @property
    def more(self):
        if self.last_response is None:
            return True
        return self.last_response.page_info.has_next_page

    @property
    def cursor(self):
        if self.last_response is None:
            return None
        return self.last_response.edges[-1].cursor

    def update_variables(self) -> None:
        self.variables.update({"cursor": self.cursor})

    def convert_objects(self) -> List["ArtifactCollection"]:
        return [
            ArtifactCollection(
                client=self.client,
                entity=self.entity,
                project=self.project,
                name=node.name,
                type=self.type_name,
            )
            for edge in self.last_response.edges
            if (node := edge.node)
        ]


class ArtifactCollection:
    def __init__(
        self,
        client: Client,
        entity: str,
        project: str,
        name: str,
        type: str,
        organization: Optional[str] = None,
        attrs: Optional[Mapping[str, Any]] = None,
        is_sequence: Optional[bool] = None,
    ):
        self.client = client
        self.entity = entity
        self.project = project
        self._name = validate_artifact_name(name)
        self._saved_name = name
        self._type = type
        self._saved_type = type
        self._attrs = attrs
        if is_sequence is not None:
            self._is_sequence = is_sequence
        is_loaded = attrs is not None and is_sequence is not None
        if not is_loaded:
            self.load()
        self._aliases = [a["node"]["alias"] for a in self._attrs["aliases"]["edges"]]
        self._description = self._attrs["description"]
        self._created_at = self._attrs["createdAt"]
        self._tags = [a["node"]["name"] for a in self._attrs["tags"]["edges"]]
        self._saved_tags = copy(self._tags)
        self.organization = organization

    @property
    def id(self) -> str:
        return self._attrs["id"]

    @normalize_exceptions
    def artifacts(self, per_page: int = 50) -> "Artifacts":
        """Artifacts."""
        return Artifacts(
            client=self.client,
            entity=self.entity,
            project=self.project,
            collection_name=self._saved_name,
            type=self._saved_type,
            per_page=per_page,
        )

    @property
    def aliases(self) -> List[str]:
        """Artifact Collection Aliases."""
        return self._aliases

    @property
    def created_at(self) -> str:
        return self._created_at

    def load(self):
        if server_supports_artifact_collections_gql_edges(self.client):
            rename_fields = None
        else:
            rename_fields = {"artifactCollection": "artifactSequence"}

        response = self.client.execute(
            gql_compat(PROJECT_ARTIFACT_COLLECTION_GQL, rename_fields=rename_fields),
            variable_values={
                "entityName": self.entity,
                "projectName": self.project,
                "artifactTypeName": self._saved_type,
                "artifactCollectionName": self._saved_name,
            },
        )

        result = ProjectArtifactCollection.model_validate(response)

        if not (
            result.project
            and (proj := result.project)
            and (type_ := proj.artifact_type)
            and (collection := type_.artifact_collection)
        ):
            raise ValueError(f"Could not find artifact type {self._saved_type}")

        sequence = type_.artifact_sequence
        self._is_sequence = (
            sequence is not None
        ) and sequence.typename__ == SOURCE_ARTIFACT_COLLECTION_TYPE

        if self._attrs is None:
            self._attrs = collection.model_dump(exclude_unset=True)
        return self._attrs

    def change_type(self, new_type: str) -> None:
        """Deprecated, change type directly with `save` instead."""
        deprecate.deprecate(
            field_name=Deprecated.artifact_collection__change_type,
            warning_message="ArtifactCollection.change_type(type) is deprecated, use ArtifactCollection.save() instead.",
        )

        if not self.is_sequence():
            raise ValueError("Artifact collection needs to be a sequence")
        termlog(
            f"Changing artifact collection type of {self._saved_type} to {new_type}"
        )
        self.client.execute(
            gql(MOVE_ARTIFACT_COLLECTION_GQL),
            variable_values={
                "artifactSequenceID": self.id,
                "destinationArtifactTypeName": new_type,
            },
        )
        self._saved_type = new_type
        self._type = new_type

    def is_sequence(self) -> bool:
        """Return whether the artifact collection is a sequence."""
        return self._is_sequence

    @normalize_exceptions
    def delete(self) -> None:
        """Delete the entire artifact collection."""
        self.client.execute(
            gql(
                DELETE_ARTIFACT_SEQUENCE_GQL
                if self.is_sequence()
                else DELETE_ARTIFACT_PORTFOLIO_GQL
            ),
            variable_values={"id": self.id},
        )

    @property
    def description(self) -> str:
        """A description of the artifact collection."""
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description

    @property
    def tags(self) -> List[str]:
        """The tags associated with the artifact collection."""
        return self._tags

    @tags.setter
    def tags(self, tags: List[str]) -> None:
        if any(not re.match(r"^[-\w]+([ ]+[-\w]+)*$", tag) for tag in tags):
            raise ValueError(
                "Tags must only contain alphanumeric characters or underscores separated by spaces or hyphens"
            )
        self._tags = tags

    @property
    def name(self) -> str:
        """The name of the artifact collection."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = validate_artifact_name(name)

    @property
    def type(self):
        """The type of the artifact collection."""
        return self._type

    @type.setter
    def type(self, type: List[str]) -> None:
        if not self.is_sequence():
            raise ValueError(
                "Type can only be changed if the artifact collection is a sequence."
            )
        self._type = type

    def _update_collection(self) -> None:
        self.client.execute(
            gql(
                UPDATE_ARTIFACT_SEQUENCE_GQL
                if self.is_sequence()
                else UPDATE_ARTIFACT_PORTFOLIO_GQL
            ),
            variable_values={
                "id": self.id,
                "name": self.name,
                "description": self.description,
            },
        )
        self._saved_name = self._name

    def _update_collection_type(self) -> None:
        self.client.execute(
            gql(MOVE_ARTIFACT_COLLECTION_GQL),
            variable_values={
                "artifactSequenceID": self.id,
                "destinationArtifactTypeName": self.type,
            },
        )
        self._saved_type = self._type

    def _add_tags(self, tags_to_add: Iterable[str]) -> None:
        self.client.execute(
            gql(CREATE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL),
            variable_values={
                "entityName": self.entity,
                "projectName": self.project,
                "artifactCollectionName": self._saved_name,
                "tags": [{"tagName": tag} for tag in tags_to_add],
            },
        )

    def _delete_tags(self, tags_to_delete: Iterable[str]) -> None:
        self.client.execute(
            gql(DELETE_ARTIFACT_COLLECTION_TAG_ASSIGNMENTS_GQL),
            variable_values={
                "entityName": self.entity,
                "projectName": self.project,
                "artifactCollectionName": self._saved_name,
                "tags": [{"tagName": tag} for tag in tags_to_delete],
            },
        )

    def save(self) -> None:
        """Persist any changes made to the artifact collection."""
        self._update_collection()

        if self.is_sequence() and (self._saved_type != self._type):
            self._update_collection_type()

        current_tags = set(self._tags)
        saved_tags = set(self._saved_tags)
        if tags_to_add := (current_tags - saved_tags):
            self._add_tags(tags_to_add)
        if tags_to_delete := (saved_tags - current_tags):
            self._delete_tags(tags_to_delete)
        self._saved_tags = copy(self._tags)

    def __repr__(self) -> str:
        return f"<ArtifactCollection {self._name} ({self._type})>"


class Artifacts(SizedPaginator["wandb.Artifact"]):
    """An iterable collection of artifact versions associated with a project and optional filter.

    This is generally used indirectly via the `Api`.artifact_versions method.
    """

    last_response: Optional[ArtifactsFragment]

    def __init__(
        self,
        client: Client,
        entity: str,
        project: str,
        collection_name: str,
        type: str,
        filters: Optional[Mapping[str, Any]] = None,
        order: Optional[str] = None,
        per_page: int = 50,
        tags: Optional[Union[str, List[str]]] = None,
    ):
        self.entity = entity
        self.collection_name = collection_name
        self.type = type
        self.project = project
        self.filters = {"state": "COMMITTED"} if filters is None else filters
        self.tags = [tags] if isinstance(tags, str) else tags
        self.order = order
        variables = {
            "project": self.project,
            "entity": self.entity,
            "order": self.order,
            "type": self.type,
            "collection": self.collection_name,
            "filters": json.dumps(self.filters),
        }

        if server_supports_artifact_collections_gql_edges(client):
            rename_fields = None
        else:
            rename_fields = {"artifactCollection": "artifactSequence"}

        self.QUERY = gql_compat(
            PROJECT_ARTIFACTS_GQL,
            omit_fields=omit_artifact_fields(api=InternalApi()),
            rename_fields=rename_fields,
        )

        super().__init__(client, variables, per_page)

    @override
    def _update_response(self) -> None:
        data = self.client.execute(self.QUERY, variable_values=self.variables)
        result = ProjectArtifacts.model_validate(data)

        # Extract the inner `*Connection` result for faster/easier access.
        if not (
            (proj := result.project)
            and (type_ := proj.artifact_type)
            and (collection := type_.artifact_collection)
            and (conn := collection.artifacts)
        ):
            raise ValueError(f"Unable to parse {type(self).__name__!r} response data")

        self.last_response = ArtifactsFragment.model_validate(conn)

    @property
    def length(self) -> Optional[int]:
        if self.last_response is None:
            return None
        return self.last_response.total_count

    @property
    def more(self) -> bool:
        if self.last_response is None:
            return True
        return self.last_response.page_info.has_next_page

    @property
    def cursor(self) -> Optional[str]:
        if self.last_response is None:
            return None
        return self.last_response.edges[-1].cursor

    def convert_objects(self) -> List["wandb.Artifact"]:
        artifact_edges = (edge for edge in self.last_response.edges if edge.node)
        artifacts = (
            wandb.Artifact._from_attrs(
                entity=self.entity,
                project=self.project,
                name=f"{self.collection_name}:{edge.version}",
                attrs=edge.node.model_dump(exclude_unset=True),
                client=self.client,
            )
            for edge in artifact_edges
        )
        required_tags = set(self.tags or [])
        return [art for art in artifacts if required_tags.issubset(art.tags)]


class RunArtifacts(SizedPaginator["wandb.Artifact"]):
    last_response: Union[
        RunOutputArtifactsProjectRunOutputArtifacts,
        RunInputArtifactsProjectRunInputArtifacts,
    ]

    #: The pydantic model used to parse the (inner part of the) raw response.
    _response_cls: Type[
        Union[
            RunOutputArtifactsProjectRunOutputArtifacts,
            RunInputArtifactsProjectRunInputArtifacts,
        ]
    ]

    def __init__(
        self,
        client: Client,
        run: "Run",
        mode: Literal["logged", "used"] = "logged",
        per_page: int = 50,
    ):
        self.run = run

        if mode == "logged":
            self.run_key = "outputArtifacts"
            self.QUERY = gql_compat(
                RUN_OUTPUT_ARTIFACTS_GQL,
                omit_fields=omit_artifact_fields(api=InternalApi()),
            )
            self._response_cls = RunOutputArtifactsProjectRunOutputArtifacts
        elif mode == "used":
            self.run_key = "inputArtifacts"
            self.QUERY = gql_compat(
                RUN_INPUT_ARTIFACTS_GQL,
                omit_fields=omit_artifact_fields(api=InternalApi()),
            )
            self._response_cls = RunInputArtifactsProjectRunInputArtifacts
        else:
            raise ValueError("mode must be logged or used")

        variable_values = {
            "entity": run.entity,
            "project": run.project,
            "runName": run.id,
        }
        super().__init__(client, variable_values, per_page)

    @override
    def _update_response(self) -> None:
        data = self.client.execute(self.QUERY, variable_values=self.variables)

        # Extract the inner `*Connection` result for faster/easier access.
        inner_data = data["project"]["run"][self.run_key]
        self.last_response = self._response_cls.model_validate(inner_data)

    @property
    def length(self) -> Optional[int]:
        if self.last_response is None:
            return None
        return self.last_response.total_count

    @property
    def more(self) -> bool:
        if self.last_response is None:
            return True
        return self.last_response.page_info.has_next_page

    @property
    def cursor(self) -> Optional[str]:
        if self.last_response is None:
            return None
        return self.last_response.edges[-1].cursor

    def convert_objects(self) -> List["wandb.Artifact"]:
        return [
            wandb.Artifact._from_attrs(
                entity=node.artifact_sequence.project.entity_name,
                project=node.artifact_sequence.project.name,
                name=f"{node.artifact_sequence.name}:v{node.version_index}",
                attrs=node.model_dump(exclude_unset=True),
                client=self.client,
            )
            for r in self.last_response.edges
            if (node := r.node)
        ]


class ArtifactFiles(SizedPaginator["public.File"]):
    last_response: Optional[FilesFragment]

    def __init__(
        self,
        client: Client,
        artifact: "wandb.Artifact",
        names: Optional[Sequence[str]] = None,
        per_page: int = 50,
    ):
        self.query_via_membership = InternalApi()._check_server_feature_with_fallback(
            ServerFeature.ARTIFACT_COLLECTION_MEMBERSHIP_FILES
        )
        self.artifact = artifact

        if self.query_via_membership:
            query_str = ARTIFACT_COLLECTION_MEMBERSHIP_FILES_GQL
            variables = {
                "entityName": artifact.entity,
                "projectName": artifact.project,
                "artifactName": artifact.name.split(":")[0],
                "artifactVersionIndex": artifact.version,
                "fileNames": names,
            }
        else:
            query_str = ARTIFACT_VERSION_FILES_GQL
            variables = {
                "entityName": artifact.source_entity,
                "projectName": artifact.source_project,
                "artifactName": artifact.source_name,
                "artifactTypeName": artifact.type,
                "fileNames": names,
            }

        # The server must advertise at least SDK 0.12.21
        # to get storagePath
        if not client.version_supported("0.12.21"):
            self.QUERY = gql_compat(query_str, omit_fields={"storagePath"})
        else:
            self.QUERY = gql(query_str)

        super().__init__(client, variables, per_page)

    @override
    def _update_response(self) -> None:
        data = self.client.execute(self.QUERY, variable_values=self.variables)

        # Extract the inner `*Connection` result for faster/easier access.
        if self.query_via_membership:
            result = ArtifactCollectionMembershipFiles.model_validate(data)
            conn = result.project.artifact_collection.artifact_membership.files
        else:
            result = ArtifactVersionFiles.model_validate(data)
            conn = result.project.artifact_type.artifact.files

        if conn is None:
            raise ValueError(f"Unable to parse {type(self).__name__!r} response data")

        self.last_response = FilesFragment.model_validate(conn)

    @property
    def path(self) -> List[str]:
        return [self.artifact.entity, self.artifact.project, self.artifact.name]

    @property
    def length(self) -> int:
        return self.artifact.file_count

    @property
    def more(self) -> bool:
        if self.last_response is None:
            return True
        return self.last_response.page_info.has_next_page

    @property
    def cursor(self) -> Optional[str]:
        if self.last_response is None:
            return None
        return self.last_response.edges[-1].cursor

    def update_variables(self) -> None:
        self.variables.update({"fileLimit": self.per_page, "fileCursor": self.cursor})

    def convert_objects(self) -> List["public.File"]:
        return [
            public.File(
                client=self.client,
                attrs=node.model_dump(exclude_unset=True),
            )
            for edge in self.last_response.edges
            if (node := edge.node)
        ]

    def __repr__(self) -> str:
        path_str = "/".join(self.path)
        return f"<ArtifactFiles {path_str} ({len(self)})>"


def server_supports_artifact_collections_gql_edges(
    client: "RetryingClient", warn: bool = False
) -> bool:
    # TODO: Validate this version
    # Edges were merged into core on Mar 2, 2022: https://github.com/wandb/core/commit/81c90b29eaacfe0a96dc1ebd83c53560ca763e8b
    # CLI version was bumped to "0.12.11" on Mar 3, 2022: https://github.com/wandb/core/commit/328396fa7c89a2178d510a1be9c0d4451f350d7b
    supported = client.version_supported("0.12.11")  # edges were merged on
    if not supported and warn:
        # First local release to include the above is 0.9.50: https://github.com/wandb/local/releases/tag/0.9.50
        wandb.termwarn(
            "W&B Local Server version does not support ArtifactCollection gql edges; falling back to using legacy ArtifactSequence. Please update server to at least version 0.9.50."
        )
    return supported



from typing import Union

from pydantic import Field
from typing_extensions import Annotated

from wandb._pydantic import GQLBase
from wandb.automations._generated import (
    GenericWebhookIntegrationFields,
    SlackIntegrationFields,
)


class SlackIntegration(SlackIntegrationFields):
    team_name: str
    """The name of the Slack workspace (not the W&B team) that this integration is associated with."""

    channel_name: str
    """The name of the Slack channel that this integration will post messages to."""


class WebhookIntegration(GenericWebhookIntegrationFields):
    name: str
    """The name of this webhook integration."""

    url_endpoint: str
    """The URL that this webhook will POST events to."""


Integration = Annotated[
    Union[SlackIntegration, WebhookIntegration],
    Field(discriminator="typename__"),
]


# For parsing integration instances from paginated responses
class _IntegrationEdge(GQLBase):
    cursor: str
    node: Integration


__all__ = [
    "Integration",
    "SlackIntegration",
    "WebhookIntegration",
]



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase

from .fragments import ProjectConnectionFields


class GetAutomationsByEntity(GQLBase):
    search_scope: Optional[GetAutomationsByEntitySearchScope] = Field(
        alias="searchScope"
    )


class GetAutomationsByEntitySearchScope(GQLBase):
    projects: Optional[ProjectConnectionFields]


GetAutomationsByEntity.model_rebuild()
GetAutomationsByEntitySearchScope.model_rebuild()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

from __future__ import annotations

from typing import Optional

from wandb._pydantic import GQLBase

from .fragments import GenericWebhookIntegrationConnectionFields


class GenericWebhookIntegrationsByEntity(GQLBase):
    entity: Optional[GenericWebhookIntegrationsByEntityEntity]


class GenericWebhookIntegrationsByEntityEntity(GQLBase):
    integrations: Optional[GenericWebhookIntegrationConnectionFields]


GenericWebhookIntegrationsByEntity.model_rebuild()
GenericWebhookIntegrationsByEntityEntity.model_rebuild()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/automations/

__all__ = [
    "CREATE_AUTOMATION_GQL",
    "CREATE_GENERIC_WEBHOOK_INTEGRATION_GQL",
    "DELETE_AUTOMATION_GQL",
    "GENERIC_WEBHOOK_INTEGRATIONS_BY_ENTITY_GQL",
    "GET_AUTOMATIONS_BY_ENTITY_GQL",
    "GET_AUTOMATIONS_GQL",
    "INTEGRATIONS_BY_ENTITY_GQL",
    "SLACK_INTEGRATIONS_BY_ENTITY_GQL",
    "UPDATE_AUTOMATION_GQL",
]

GET_AUTOMATIONS_GQL = """
query GetAutomations($cursor: String, $perPage: Int) {
  searchScope: viewer {
    projects(after: $cursor, first: $perPage) {
      ...ProjectConnectionFields
    }
  }
}

fragment ArtifactPortfolioScopeFields on ArtifactPortfolio {
  __typename
  id
  name
}

fragment ArtifactSequenceScopeFields on ArtifactSequence {
  __typename
  id
  name
}

fragment FilterEventFields on FilterEventTriggeringCondition {
  __typename
  eventType
  filter
}

fragment GenericWebhookActionFields on GenericWebhookTriggeredAction {
  __typename
  integration {
    __typename
    ...GenericWebhookIntegrationFields
  }
  requestPayload
}

fragment GenericWebhookIntegrationFields on GenericWebhookIntegration {
  __typename
  id
  name
  urlEndpoint
}

fragment NoOpActionFields on NoOpTriggeredAction {
  __typename
  noOp
}

fragment NotificationActionFields on NotificationTriggeredAction {
  __typename
  integration {
    __typename
    ...SlackIntegrationFields
  }
  title
  message
  severity
}

fragment PageInfoFields on PageInfo {
  endCursor
  hasNextPage
}

fragment ProjectConnectionFields on ProjectConnection {
  __typename
  pageInfo {
    ...PageInfoFields
  }
  edges {
    cursor
    node {
      triggers {
        ...TriggerFields
      }
    }
  }
}

fragment ProjectScopeFields on Project {
  __typename
  id
  name
}

fragment QueueJobActionFields on QueueJobTriggeredAction {
  __typename
  queue {
    id
    name
  }
  template
}

fragment SlackIntegrationFields on SlackIntegration {
  __typename
  id
  teamName
  channelName
}

fragment TriggerFields on Trigger {
  __typename
  id
  createdAt
  updatedAt
  name
  description
  enabled
  scope {
    __typename
    ...ProjectScopeFields
    ...ArtifactPortfolioScopeFields
    ...ArtifactSequenceScopeFields
  }
  event: triggeringCondition {
    __typename
    ...FilterEventFields
  }
  action: triggeredAction {
    __typename
    ...QueueJobActionFields
    ...NotificationActionFields
    ...GenericWebhookActionFields
    ...NoOpActionFields
  }
}
"""

GET_AUTOMATIONS_BY_ENTITY_GQL = """
query GetAutomationsByEntity($entityName: String!, $cursor: String, $perPage: Int) {
  searchScope: entity(name: $entityName) {
    projects(after: $cursor, first: $perPage) {
      ...ProjectConnectionFields
    }
  }
}

fragment ArtifactPortfolioScopeFields on ArtifactPortfolio {
  __typename
  id
  name
}

fragment ArtifactSequenceScopeFields on ArtifactSequence {
  __typename
  id
  name
}

fragment FilterEventFields on FilterEventTriggeringCondition {
  __typename
  eventType
  filter
}

fragment GenericWebhookActionFields on GenericWebhookTriggeredAction {
  __typename
  integration {
    __typename
    ...GenericWebhookIntegrationFields
  }
  requestPayload
}

fragment GenericWebhookIntegrationFields on GenericWebhookIntegration {
  __typename
  id
  name
  urlEndpoint
}

fragment NoOpActionFields on NoOpTriggeredAction {
  __typename
  noOp
}

fragment NotificationActionFields on NotificationTriggeredAction {
  __typename
  integration {
    __typename
    ...SlackIntegrationFields
  }
  title
  message
  severity
}

fragment PageInfoFields on PageInfo {
  endCursor
  hasNextPage
}

fragment ProjectConnectionFields on ProjectConnection {
  __typename
  pageInfo {
    ...PageInfoFields
  }
  edges {
    cursor
    node {
      triggers {
        ...TriggerFields
      }
    }
  }
}

fragment ProjectScopeFields on Project {
  __typename
  id
  name
}

fragment QueueJobActionFields on QueueJobTriggeredAction {
  __typename
  queue {
    id
    name
  }
  template
}

fragment SlackIntegrationFields on SlackIntegration {
  __typename
  id
  teamName
  channelName
}

fragment TriggerFields on Trigger {
  __typename
  id
  createdAt
  updatedAt
  name
  description
  enabled
  scope {
    __typename
    ...ProjectScopeFields
    ...ArtifactPortfolioScopeFields
    ...ArtifactSequenceScopeFields
  }
  event: triggeringCondition {
    __typename
    ...FilterEventFields
  }
  action: triggeredAction {
    __typename
    ...QueueJobActionFields
    ...NotificationActionFields
    ...GenericWebhookActionFields
    ...NoOpActionFields
  }
}
"""

CREATE_AUTOMATION_GQL = """
mutation CreateAutomation($params: CreateFilterTriggerInput!) {
  result: createFilterTrigger(input: $params) {
    ...CreateAutomationResult
  }
}

fragment ArtifactPortfolioScopeFields on ArtifactPortfolio {
  __typename
  id
  name
}

fragment ArtifactSequenceScopeFields on ArtifactSequence {
  __typename
  id
  name
}

fragment CreateAutomationResult on CreateFilterTriggerPayload {
  __typename
  trigger {
    ...TriggerFields
  }
}

fragment FilterEventFields on FilterEventTriggeringCondition {
  __typename
  eventType
  filter
}

fragment GenericWebhookActionFields on GenericWebhookTriggeredAction {
  __typename
  integration {
    __typename
    ...GenericWebhookIntegrationFields
  }
  requestPayload
}

fragment GenericWebhookIntegrationFields on GenericWebhookIntegration {
  __typename
  id
  name
  urlEndpoint
}

fragment NoOpActionFields on NoOpTriggeredAction {
  __typename
  noOp
}

fragment NotificationActionFields on NotificationTriggeredAction {
  __typename
  integration {
    __typename
    ...SlackIntegrationFields
  }
  title
  message
  severity
}

fragment ProjectScopeFields on Project {
  __typename
  id
  name
}

fragment QueueJobActionFields on QueueJobTriggeredAction {
  __typename
  queue {
    id
    name
  }
  template
}

fragment SlackIntegrationFields on SlackIntegration {
  __typename
  id
  teamName
  channelName
}

fragment TriggerFields on Trigger {
  __typename
  id
  createdAt
  updatedAt
  name
  description
  enabled
  scope {
    __typename
    ...ProjectScopeFields
    ...ArtifactPortfolioScopeFields
    ...ArtifactSequenceScopeFields
  }
  event: triggeringCondition {
    __typename
    ...FilterEventFields
  }
  action: triggeredAction {
    __typename
    ...QueueJobActionFields
    ...NotificationActionFields
    ...GenericWebhookActionFields
    ...NoOpActionFields
  }
}
"""

UPDATE_AUTOMATION_GQL = """
mutation UpdateAutomation($params: UpdateFilterTriggerInput!) {
  result: updateFilterTrigger(input: $params) {
    ...UpdateAutomationResult
  }
}

fragment ArtifactPortfolioScopeFields on ArtifactPortfolio {
  __typename
  id
  name
}

fragment ArtifactSequenceScopeFields on ArtifactSequence {
  __typename
  id
  name
}

fragment FilterEventFields on FilterEventTriggeringCondition {
  __typename
  eventType
  filter
}

fragment GenericWebhookActionFields on GenericWebhookTriggeredAction {
  __typename
  integration {
    __typename
    ...GenericWebhookIntegrationFields
  }
  requestPayload
}

fragment GenericWebhookIntegrationFields on GenericWebhookIntegration {
  __typename
  id
  name
  urlEndpoint
}

fragment NoOpActionFields on NoOpTriggeredAction {
  __typename
  noOp
}

fragment NotificationActionFields on NotificationTriggeredAction {
  __typename
  integration {
    __typename
    ...SlackIntegrationFields
  }
  title
  message
  severity
}

fragment ProjectScopeFields on Project {
  __typename
  id
  name
}

fragment QueueJobActionFields on QueueJobTriggeredAction {
  __typename
  queue {
    id
    name
  }
  template
}

fragment SlackIntegrationFields on SlackIntegration {
  __typename
  id
  teamName
  channelName
}

fragment TriggerFields on Trigger {
  __typename
  id
  createdAt
  updatedAt
  name
  description
  enabled
  scope {
    __typename
    ...ProjectScopeFields
    ...ArtifactPortfolioScopeFields
    ...ArtifactSequenceScopeFields
  }
  event: triggeringCondition {
    __typename
    ...FilterEventFields
  }
  action: triggeredAction {
    __typename
    ...QueueJobActionFields
    ...NotificationActionFields
    ...GenericWebhookActionFields
    ...NoOpActionFields
  }
}

fragment UpdateAutomationResult on UpdateFilterTriggerPayload {
  __typename
  trigger {
    ...TriggerFields
  }
}
"""

DELETE_AUTOMATION_GQL = """
mutation DeleteAutomation($id: ID!) {
  result: deleteTrigger(input: {triggerID: $id}) {
    ...DeleteAutomationResult
  }
}

fragment DeleteAutomationResult on DeleteTriggerPayload {
  __typename
  success
}
"""

INTEGRATIONS_BY_ENTITY_GQL = """
query IntegrationsByEntity($entityName: String!, $cursor: String, $perPage: Int) {
  entity(name: $entityName) {
    integrations(after: $cursor, first: $perPage) {
      ...IntegrationConnectionFields
    }
  }
}

fragment GenericWebhookIntegrationFields on GenericWebhookIntegration {
  __typename
  id
  name
  urlEndpoint
}

fragment IntegrationConnectionFields on IntegrationConnection {
  __typename
  pageInfo {
    ...PageInfoFields
  }
  edges {
    cursor
    node {
      __typename
      ...SlackIntegrationFields
      ...GenericWebhookIntegrationFields
    }
  }
}

fragment PageInfoFields on PageInfo {
  endCursor
  hasNextPage
}

fragment SlackIntegrationFields on SlackIntegration {
  __typename
  id
  teamName
  channelName
}
"""

SLACK_INTEGRATIONS_BY_ENTITY_GQL = """
query SlackIntegrationsByEntity($entityName: String!, $cursor: String, $perPage: Int) {
  entity(name: $entityName) {
    integrations(after: $cursor, first: $perPage) {
      ...SlackIntegrationConnectionFields
    }
  }
}

fragment PageInfoFields on PageInfo {
  endCursor
  hasNextPage
}

fragment SlackIntegrationConnectionFields on IntegrationConnection {
  __typename
  pageInfo {
    ...PageInfoFields
  }
  edges {
    cursor
    node {
      __typename
      ...SlackIntegrationFields
    }
  }
}

fragment SlackIntegrationFields on SlackIntegration {
  __typename
  id
  teamName
  channelName
}
"""

GENERIC_WEBHOOK_INTEGRATIONS_BY_ENTITY_GQL = """
query GenericWebhookIntegrationsByEntity($entityName: String!, $cursor: String, $perPage: Int) {
  entity(name: $entityName) {
    integrations(after: $cursor, first: $perPage) {
      ...GenericWebhookIntegrationConnectionFields
    }
  }
}

fragment GenericWebhookIntegrationConnectionFields on IntegrationConnection {
  __typename
  pageInfo {
    ...PageInfoFields
  }
  edges {
    cursor
    node {
      __typename
      ...GenericWebhookIntegrationFields
    }
  }
}

fragment GenericWebhookIntegrationFields on GenericWebhookIntegration {
  __typename
  id
  name
  urlEndpoint
}

fragment PageInfoFields on PageInfo {
  endCursor
  hasNextPage
}
"""

CREATE_GENERIC_WEBHOOK_INTEGRATION_GQL = """
mutation CreateGenericWebhookIntegration($params: CreateGenericWebhookIntegrationInput!) {
  createGenericWebhookIntegration(input: $params) {
    integration {
      __typename
      ...GenericWebhookIntegrationFields
    }
  }
}

fragment GenericWebhookIntegrationFields on GenericWebhookIntegration {
  __typename
  id
  name
  urlEndpoint
}
"""



from __future__ import annotations

from typing import TYPE_CHECKING

import wandb
from wandb.proto.wandb_deprecated import DEPRECATED_FEATURES
from wandb.sdk.lib import telemetry

# Necessary to break import cycle.
if TYPE_CHECKING:
    from wandb import wandb_run


def deprecate(
    field_name: DEPRECATED_FEATURES,
    warning_message: str,
    run: wandb_run.Run | None = None,
) -> None:
    """Warn the user that a feature has been deprecated.

    If a run is provided, the given field on its telemetry is updated.
    Otherwise, the global run is used.

    Args:
        field_name: The field on the Deprecated proto for this deprecation.
        warning_message: The message to display to the user.
        run: The run whose telemetry to update.
    """
    _run = run or wandb.run
    with telemetry.context(run=_run) as tel:
        setattr(tel.deprecated, field_name, True)

    wandb.termwarn(warning_message, repeat=False)



import base64
import binascii
import codecs
import datetime
import logging
import os

import wandb
from wandb import util
from wandb.sdk.lib import runid

from . import _dtypes
from ._private import MEDIA_TMP
from .base_types.media import Media, _numpy_arrays_to_lists
from .base_types.wb_value import WBValue
from .utils import _json_helper


class _TableLinkMixin:
    def set_table(self, table):
        self._table = table


class _TableKey(str, _TableLinkMixin):
    def set_table(self, table, col_name):
        assert col_name in table.columns
        self._table = table
        self._col_name = col_name


class _TableIndex(int, _TableLinkMixin):
    def get_row(self):
        row = {}
        if self._table:
            row = {
                c: self._table.data[self][i] for i, c in enumerate(self._table.columns)
            }

        return row


class _PrimaryKeyType(_dtypes.Type):
    name = "primaryKey"
    legacy_names = ["wandb.TablePrimaryKey"]

    def assign_type(self, wb_type=None):
        if isinstance(wb_type, _dtypes.StringType) or isinstance(
            wb_type, _PrimaryKeyType
        ):
            return self
        return _dtypes.InvalidType()

    @classmethod
    def from_obj(cls, py_obj):
        if not isinstance(py_obj, _TableKey):
            raise TypeError("py_obj must be a wandb.Table")
        else:
            return cls()


class _ForeignKeyType(_dtypes.Type):
    name = "foreignKey"
    legacy_names = ["wandb.TableForeignKey"]
    types = [_TableKey]

    def __init__(self, table, col_name):
        assert isinstance(table, Table)
        assert isinstance(col_name, str)
        assert col_name in table.columns
        self.params.update({"table": table, "col_name": col_name})

    def assign_type(self, wb_type=None):
        if isinstance(wb_type, _dtypes.StringType):
            return self
        elif (
            isinstance(wb_type, _ForeignKeyType)
            and id(self.params["table"]) == id(wb_type.params["table"])
            and self.params["col_name"] == wb_type.params["col_name"]
        ):
            return self

        return _dtypes.InvalidType()

    @classmethod
    def from_obj(cls, py_obj):
        if not isinstance(py_obj, _TableKey):
            raise TypeError("py_obj must be a _TableKey")
        else:
            return cls(py_obj._table, py_obj._col_name)

    def to_json(self, artifact=None):
        res = super().to_json(artifact)
        if artifact is not None:
            table_name = f"media/tables/t_{runid.generate_id()}"
            entry = artifact.add(self.params["table"], table_name)
            res["params"]["table"] = entry.path
        else:
            raise AssertionError(
                "_ForeignKeyType does not support serialization without an artifact"
            )
        return res

    @classmethod
    def from_json(
        cls,
        json_dict,
        artifact,
    ):
        table = None
        col_name = None
        if artifact is None:
            raise AssertionError(
                "_ForeignKeyType does not support deserialization without an artifact"
            )
        else:
            table = artifact.get(json_dict["params"]["table"])
            col_name = json_dict["params"]["col_name"]

        if table is None:
            raise AssertionError("Unable to deserialize referenced table")

        return cls(table, col_name)


class _ForeignIndexType(_dtypes.Type):
    name = "foreignIndex"
    legacy_names = ["wandb.TableForeignIndex"]
    types = [_TableIndex]

    def __init__(self, table):
        assert isinstance(table, Table)
        self.params.update({"table": table})

    def assign_type(self, wb_type=None):
        if isinstance(wb_type, _dtypes.NumberType):
            return self
        elif isinstance(wb_type, _ForeignIndexType) and id(self.params["table"]) == id(
            wb_type.params["table"]
        ):
            return self

        return _dtypes.InvalidType()

    @classmethod
    def from_obj(cls, py_obj):
        if not isinstance(py_obj, _TableIndex):
            raise TypeError("py_obj must be a _TableIndex")
        else:
            return cls(py_obj._table)

    def to_json(self, artifact=None):
        res = super().to_json(artifact)
        if artifact is not None:
            table_name = f"media/tables/t_{runid.generate_id()}"
            entry = artifact.add(self.params["table"], table_name)
            res["params"]["table"] = entry.path
        else:
            raise AssertionError(
                "_ForeignIndexType does not support serialization without an artifact"
            )
        return res

    @classmethod
    def from_json(
        cls,
        json_dict,
        artifact,
    ):
        table = None
        if artifact is None:
            raise AssertionError(
                "_ForeignIndexType does not support deserialization without an artifact"
            )
        else:
            table = artifact.get(json_dict["params"]["table"])

        if table is None:
            raise AssertionError("Unable to deserialize referenced table")

        return cls(table)


class Table(Media):
    """The Table class used to display and analyze tabular data.

    Unlike traditional spreadsheets, Tables support numerous types of data:
    scalar values, strings, numpy arrays, and most subclasses of `wandb.data_types.Media`.
    This means you can embed `Images`, `Video`, `Audio`, and other sorts of rich, annotated media
    directly in Tables, alongside other traditional scalar values.

    This class is the primary class used to generate the Table Visualizer
    in the UI: https://docs.wandb.ai/guides/data-vis/tables.

    Args:
        columns: (List[str]) Names of the columns in the table.
            Defaults to ["Input", "Output", "Expected"].
        data: (List[List[any]]) 2D row-oriented array of values.
        dataframe: (pandas.DataFrame) DataFrame object used to create the table.
            When set, `data` and `columns` arguments are ignored.
        optional: (Union[bool,List[bool]]) Determines if `None` values are allowed. Default to True
            - If a singular bool value, then the optionality is enforced for all
            columns specified at construction time
            - If a list of bool values, then the optionality is applied to each
            column - should be the same length as `columns`
            applies to all columns. A list of bool values applies to each respective column.
        allow_mixed_types: (bool) Determines if columns are allowed to have mixed types
            (disables type validation). Defaults to False
    """

    MAX_ROWS = 10000
    MAX_ARTIFACT_ROWS = 200000
    _MAX_EMBEDDING_DIMENSIONS = 150
    _log_type = "table"

    def __init__(
        self,
        columns=None,
        data=None,
        rows=None,
        dataframe=None,
        dtype=None,
        optional=True,
        allow_mixed_types=False,
    ):
        """Initializes a Table object.

        The rows is available for legacy reasons and should not be used.
        The Table class uses data to mimic the Pandas API.
        """
        super().__init__()
        self._pk_col = None
        self._fk_cols = set()
        if allow_mixed_types:
            dtype = _dtypes.AnyType

        # This is kept for legacy reasons (tss: personally, I think we should remove this)
        if columns is None:
            columns = ["Input", "Output", "Expected"]

        # Explicit dataframe option
        if dataframe is not None:
            self._init_from_dataframe(dataframe, columns, optional, dtype)
        else:
            # Expected pattern
            if data is not None:
                if util.is_numpy_array(data):
                    self._init_from_ndarray(data, columns, optional, dtype)
                elif util.is_pandas_data_frame(data):
                    self._init_from_dataframe(data, columns, optional, dtype)
                else:
                    self._init_from_list(data, columns, optional, dtype)

            # legacy
            elif rows is not None:
                self._init_from_list(rows, columns, optional, dtype)

            # Default empty case
            else:
                self._init_from_list([], columns, optional, dtype)

    @staticmethod
    def _assert_valid_columns(columns):
        valid_col_types = [str, int]
        assert isinstance(columns, list), "columns argument expects a `list` object"
        assert len(columns) == 0 or all(
            [type(col) in valid_col_types for col in columns]
        ), "columns argument expects list of strings or ints"

    def _init_from_list(self, data, columns, optional=True, dtype=None):
        assert isinstance(data, list), "data argument expects a `list` object"
        self.data = []
        self._assert_valid_columns(columns)
        self.columns = columns
        self._make_column_types(dtype, optional)
        for row in data:
            self.add_data(*row)

    def _init_from_ndarray(self, ndarray, columns, optional=True, dtype=None):
        assert util.is_numpy_array(
            ndarray
        ), "ndarray argument expects a `numpy.ndarray` object"
        self.data = []
        self._assert_valid_columns(columns)
        self.columns = columns
        self._make_column_types(dtype, optional)
        for row in ndarray:
            self.add_data(*row)

    def _init_from_dataframe(self, dataframe, columns, optional=True, dtype=None):
        assert util.is_pandas_data_frame(
            dataframe
        ), "dataframe argument expects a `pandas.core.frame.DataFrame` object"
        self.data = []
        columns = list(dataframe.columns)
        self._assert_valid_columns(columns)
        self.columns = columns
        self._make_column_types(dtype, optional)
        for row in range(len(dataframe)):
            self.add_data(*tuple(dataframe[col].values[row] for col in self.columns))

    def _make_column_types(self, dtype=None, optional=True):
        if dtype is None:
            dtype = _dtypes.UnknownType()

        if optional.__class__ is not list:
            optional = [optional for _ in range(len(self.columns))]

        if dtype.__class__ is not list:
            dtype = [dtype for _ in range(len(self.columns))]

        self._column_types = _dtypes.TypedDictType({})
        for col_name, opt, dt in zip(self.columns, optional, dtype):
            self.cast(col_name, dt, opt)

    def cast(self, col_name, dtype, optional=False):
        """Casts a column to a specific data type.

        This can be one of the normal python classes, an internal W&B type, or an
        example object, like an instance of wandb.Image or wandb.Classes.

        Args:
            col_name: (str) - The name of the column to cast.
            dtype: (class, wandb.wandb_sdk.interface._dtypes.Type, any) - The target dtype.
            optional: (bool) - If the column should allow Nones.
        """
        assert col_name in self.columns

        wbtype = _dtypes.TypeRegistry.type_from_dtype(dtype)

        if optional:
            wbtype = _dtypes.OptionalType(wbtype)

        # Cast each value in the row, raising an error if there are invalid entries.
        col_ndx = self.columns.index(col_name)
        for row in self.data:
            result_type = wbtype.assign(row[col_ndx])
            if isinstance(result_type, _dtypes.InvalidType):
                raise TypeError(
                    "Existing data {}, of type {} cannot be cast to {}".format(
                        row[col_ndx],
                        _dtypes.TypeRegistry.type_of(row[col_ndx]),
                        wbtype,
                    )
                )
            wbtype = result_type

        # Assert valid options
        is_pk = isinstance(wbtype, _PrimaryKeyType)
        is_fk = isinstance(wbtype, _ForeignKeyType)
        is_fi = isinstance(wbtype, _ForeignIndexType)
        if is_pk or is_fk or is_fi:
            assert (
                not optional
            ), "Primary keys, foreign keys, and foreign indexes cannot be optional."

        if (is_fk or is_fk) and id(wbtype.params["table"]) == id(self):
            raise AssertionError("Cannot set a foreign table reference to same table.")

        if is_pk:
            assert (
                self._pk_col is None
            ), "Cannot have multiple primary keys - {} is already set as the primary key.".format(
                self._pk_col
            )

        # Update the column type
        self._column_types.params["type_map"][col_name] = wbtype

        # Wrap the data if needed
        self._update_keys()
        return wbtype

    def __ne__(self, other):
        return not self.__eq__(other)

    def _eq_debug(self, other, should_assert=False):
        eq = isinstance(other, Table)
        assert not should_assert or eq, "Found type {}, expected {}".format(
            other.__class__, Table
        )
        eq = eq and len(self.data) == len(other.data)
        assert not should_assert or eq, "Found {} rows, expected {}".format(
            len(other.data), len(self.data)
        )
        eq = eq and self.columns == other.columns
        assert not should_assert or eq, "Found columns {}, expected {}".format(
            other.columns, self.columns
        )
        eq = eq and self._column_types == other._column_types
        assert (
            not should_assert or eq
        ), "Found column type {}, expected column type {}".format(
            other._column_types, self._column_types
        )
        if eq:
            for row_ndx in range(len(self.data)):
                for col_ndx in range(len(self.data[row_ndx])):
                    _eq = self.data[row_ndx][col_ndx] == other.data[row_ndx][col_ndx]
                    # equal if all are equal
                    if util.is_numpy_array(_eq):
                        _eq = ((_eq * -1) + 1).sum() == 0
                    eq = eq and _eq
                    assert (
                        not should_assert or eq
                    ), "Unequal data at row_ndx {} col_ndx {}: found {}, expected {}".format(
                        row_ndx,
                        col_ndx,
                        other.data[row_ndx][col_ndx],
                        self.data[row_ndx][col_ndx],
                    )
                    if not eq:
                        return eq
        return eq

    def __eq__(self, other):
        return self._eq_debug(other)

    def add_row(self, *row):
        """Deprecated; use add_data instead."""
        logging.warning("add_row is deprecated, use add_data")
        self.add_data(*row)

    def add_data(self, *data):
        """Adds a new row of data to the table. The maximum amount of rows in a table is determined by `wandb.Table.MAX_ARTIFACT_ROWS`.

        The length of the data should match the length of the table column.
        """
        if len(data) != len(self.columns):
            raise ValueError(
                "This table expects {} columns: {}, found {}".format(
                    len(self.columns), self.columns, len(data)
                )
            )

        # Special case to pre-emptively cast a column as a key.
        # Needed as String.assign(Key) is invalid
        for ndx, item in enumerate(data):
            if isinstance(item, _TableLinkMixin):
                self.cast(
                    self.columns[ndx],
                    _dtypes.TypeRegistry.type_of(item),
                    optional=False,
                )

        # Update the table's column types
        result_type = self._get_updated_result_type(data)
        self._column_types = result_type

        # rows need to be mutable
        if isinstance(data, tuple):
            data = list(data)
        # Add the new data
        self.data.append(data)

        # Update the wrapper values if needed
        self._update_keys(force_last=True)

    def _get_updated_result_type(self, row):
        """Returns the updated result type based on the inputted row.

        Raises:
            TypeError: if the assignment is invalid.
        """
        incoming_row_dict = {
            col_key: row[ndx] for ndx, col_key in enumerate(self.columns)
        }
        current_type = self._column_types
        result_type = current_type.assign(incoming_row_dict)
        if isinstance(result_type, _dtypes.InvalidType):
            raise TypeError(
                "Data row contained incompatible types:\n{}".format(
                    current_type.explain(incoming_row_dict)
                )
            )
        return result_type

    def _to_table_json(self, max_rows=None, warn=True):
        # separate this method for easier testing
        if max_rows is None:
            max_rows = Table.MAX_ROWS
        n_rows = len(self.data)
        if n_rows > max_rows and warn:
            # NOTE: Never raises for reinit="create_new" runs.
            #   Since this is called by bind_to_run(), this can be fixed by
            #   propagating the run. It cannot be fixed for to_json() calls
            #   that are given an artifact, other than by deferring to singleton
            #   settings.
            if wandb.run and (
                wandb.run.settings.table_raise_on_max_row_limit_exceeded
                or wandb.run.settings.strict
            ):
                raise ValueError(
                    f"Table row limit exceeded: table has {n_rows} rows, limit is {max_rows}. "
                    f"To increase the maximum number of allowed rows in a wandb.Table, override "
                    f"the limit with `wandb.Table.MAX_ARTIFACT_ROWS = X` and try again. Note: "
                    f"this may cause slower queries in the W&B UI."
                )
            logging.warning(f"Truncating wandb.Table object to {max_rows} rows.")
        return {"columns": self.columns, "data": self.data[:max_rows]}

    def bind_to_run(self, *args, **kwargs):
        # We set `warn=False` since Tables will now always be logged to both
        # files and artifacts. The file limit will never practically matter and
        # this code path will be ultimately removed. The 10k limit warning confuses
        # users given that we publicly say 200k is the limit.
        data = self._to_table_json(warn=False)
        tmp_path = os.path.join(MEDIA_TMP.name, runid.generate_id() + ".table.json")
        data = _numpy_arrays_to_lists(data)
        with codecs.open(tmp_path, "w", encoding="utf-8") as fp:
            util.json_dump_safer(data, fp)
        self._set_file(tmp_path, is_tmp=True, extension=".table.json")
        super().bind_to_run(*args, **kwargs)

    @classmethod
    def get_media_subdir(cls):
        return os.path.join("media", "table")

    @classmethod
    def from_json(cls, json_obj, source_artifact):
        data = []
        column_types = None
        np_deserialized_columns = {}
        timestamp_column_indices = set()
        if json_obj.get("column_types") is not None:
            column_types = _dtypes.TypeRegistry.type_from_dict(
                json_obj["column_types"], source_artifact
            )
            for col_name in column_types.params["type_map"]:
                col_type = column_types.params["type_map"][col_name]
                ndarray_type = None
                if isinstance(col_type, _dtypes.NDArrayType):
                    ndarray_type = col_type
                elif isinstance(col_type, _dtypes.UnionType):
                    for t in col_type.params["allowed_types"]:
                        if isinstance(t, _dtypes.NDArrayType):
                            ndarray_type = t
                        elif isinstance(t, _dtypes.TimestampType):
                            timestamp_column_indices.add(
                                json_obj["columns"].index(col_name)
                            )

                elif isinstance(col_type, _dtypes.TimestampType):
                    timestamp_column_indices.add(json_obj["columns"].index(col_name))

                if (
                    ndarray_type is not None
                    and ndarray_type._get_serialization_path() is not None
                ):
                    serialization_path = ndarray_type._get_serialization_path()
                    np = util.get_module(
                        "numpy",
                        required="Deserializing NumPy columns requires NumPy to be installed.",
                    )
                    deserialized = np.load(
                        source_artifact.get_entry(serialization_path["path"]).download()
                    )
                    np_deserialized_columns[json_obj["columns"].index(col_name)] = (
                        deserialized[serialization_path["key"]]
                    )
                    ndarray_type._clear_serialization_path()

        for r_ndx, row in enumerate(json_obj["data"]):
            row_data = []
            for c_ndx, item in enumerate(row):
                cell = item
                if c_ndx in timestamp_column_indices and isinstance(item, (int, float)):
                    cell = datetime.datetime.fromtimestamp(
                        item / 1000, tz=datetime.timezone.utc
                    )
                elif c_ndx in np_deserialized_columns:
                    cell = np_deserialized_columns[c_ndx][r_ndx]
                elif isinstance(item, dict) and "_type" in item:
                    obj = WBValue.init_from_json(item, source_artifact)
                    if obj is not None:
                        cell = obj
                row_data.append(cell)
            data.append(row_data)

        # construct Table with dtypes for each column if type information exists
        dtypes = None
        if column_types is not None:
            dtypes = [
                column_types.params["type_map"][str(col)] for col in json_obj["columns"]
            ]

        new_obj = cls(columns=json_obj["columns"], data=data, dtype=dtypes)

        if column_types is not None:
            new_obj._column_types = column_types

        new_obj._update_keys()
        return new_obj

    def to_json(self, run_or_artifact):
        json_dict = super().to_json(run_or_artifact)

        if isinstance(run_or_artifact, wandb.wandb_sdk.wandb_run.Run):
            json_dict.update(
                {
                    "_type": "table-file",
                    "ncols": len(self.columns),
                    "nrows": len(self.data),
                }
            )

        elif isinstance(run_or_artifact, wandb.Artifact):
            artifact = run_or_artifact
            mapped_data = []
            data = self._to_table_json(Table.MAX_ARTIFACT_ROWS)["data"]

            ndarray_col_ndxs = set()
            for col_ndx, col_name in enumerate(self.columns):
                col_type = self._column_types.params["type_map"][col_name]
                ndarray_type = None
                if isinstance(col_type, _dtypes.NDArrayType):
                    ndarray_type = col_type
                elif isinstance(col_type, _dtypes.UnionType):
                    for t in col_type.params["allowed_types"]:
                        if isinstance(t, _dtypes.NDArrayType):
                            ndarray_type = t

                # Do not serialize 1d arrays - these are likely embeddings and
                # will not have the same cost as higher dimensional arrays
                is_1d_array = (
                    ndarray_type is not None
                    and "shape" in ndarray_type._params
                    and isinstance(ndarray_type._params["shape"], list)
                    and len(ndarray_type._params["shape"]) == 1
                    and ndarray_type._params["shape"][0]
                    <= self._MAX_EMBEDDING_DIMENSIONS
                )
                if is_1d_array:
                    self._column_types.params["type_map"][col_name] = _dtypes.ListType(
                        _dtypes.NumberType, ndarray_type._params["shape"][0]
                    )
                elif ndarray_type is not None:
                    np = util.get_module(
                        "numpy",
                        required="Serializing NumPy requires NumPy to be installed.",
                    )
                    file_name = f"{str(col_name)}_{runid.generate_id()}.npz"
                    npz_file_name = os.path.join(MEDIA_TMP.name, file_name)
                    np.savez_compressed(
                        npz_file_name,
                        **{
                            str(col_name): self.get_column(col_name, convert_to="numpy")
                        },
                    )
                    entry = artifact.add_file(
                        npz_file_name, "media/serialized_data/" + file_name, is_tmp=True
                    )
                    ndarray_type._set_serialization_path(entry.path, str(col_name))
                    ndarray_col_ndxs.add(col_ndx)

            for row in data:
                mapped_row = []
                for ndx, v in enumerate(row):
                    if ndx in ndarray_col_ndxs:
                        mapped_row.append(None)
                    else:
                        mapped_row.append(_json_helper(v, artifact))
                mapped_data.append(mapped_row)

            json_dict.update(
                {
                    "_type": Table._log_type,
                    "columns": self.columns,
                    "data": mapped_data,
                    "ncols": len(self.columns),
                    "nrows": len(mapped_data),
                    "column_types": self._column_types.to_json(artifact),
                }
            )
        else:
            raise ValueError("to_json accepts wandb_run.Run or wandb_artifact.Artifact")

        return json_dict

    def iterrows(self):
        """Returns the table data by row, showing the index of the row and the relevant data.

        Yields:
        ------
        index : int
            The index of the row. Using this value in other W&B tables
            will automatically build a relationship between the tables
        row : List[any]
            The data of the row.
        """
        for ndx in range(len(self.data)):
            index = _TableIndex(ndx)
            index.set_table(self)
            yield index, self.data[ndx]

    def set_pk(self, col_name):
        # TODO: Docs
        assert col_name in self.columns
        self.cast(col_name, _PrimaryKeyType())

    def set_fk(self, col_name, table, table_col):
        # TODO: Docs
        assert col_name in self.columns
        assert col_name != self._pk_col
        self.cast(col_name, _ForeignKeyType(table, table_col))

    def _update_keys(self, force_last=False):
        """Updates the known key-like columns based on current column types.

        If the state has been updated since the last update, wraps the data
        appropriately in the Key classes.

        Args:
            force_last: (bool) Wraps the last column of data even if there
                are no key updates.
        """
        _pk_col = None
        _fk_cols = set()

        # Buildup the known keys from column types
        c_types = self._column_types.params["type_map"]
        for t in c_types:
            if isinstance(c_types[t], _PrimaryKeyType):
                _pk_col = t
            elif isinstance(c_types[t], _ForeignKeyType) or isinstance(
                c_types[t], _ForeignIndexType
            ):
                _fk_cols.add(t)

        # If there are updates to perform, safely update them
        has_update = _pk_col != self._pk_col or _fk_cols != self._fk_cols
        if has_update:
            # If we removed the PK
            if _pk_col is None and self._pk_col is not None:
                raise AssertionError(
                    f"Cannot unset primary key (column {self._pk_col})"
                )
            # If there is a removed FK
            if len(self._fk_cols - _fk_cols) > 0:
                raise AssertionError(
                    "Cannot unset foreign key. Attempted to unset ({})".format(
                        self._fk_cols - _fk_cols
                    )
                )

            self._pk_col = _pk_col
            self._fk_cols = _fk_cols

        # Apply updates to data only if there are update or the caller
        # requested the final row to be updated
        if has_update or force_last:
            self._apply_key_updates(not has_update)

    def _apply_key_updates(self, only_last=False):
        """Appropriately wraps the underlying data in special Key classes.

        Args:
            only_last: only apply the updates to the last row (used for performance when
            the caller knows that the only new data is the last row and no updates were
            applied to the column types)
        """
        c_types = self._column_types.params["type_map"]

        # Define a helper function which will wrap the data of a single row
        # in the appropriate class wrapper.
        def update_row(row_ndx):
            for fk_col in self._fk_cols:
                col_ndx = self.columns.index(fk_col)

                # Wrap the Foreign Keys
                if isinstance(c_types[fk_col], _ForeignKeyType) and not isinstance(
                    self.data[row_ndx][col_ndx], _TableKey
                ):
                    self.data[row_ndx][col_ndx] = _TableKey(self.data[row_ndx][col_ndx])
                    self.data[row_ndx][col_ndx].set_table(
                        c_types[fk_col].params["table"],
                        c_types[fk_col].params["col_name"],
                    )

                # Wrap the Foreign Indexes
                elif isinstance(c_types[fk_col], _ForeignIndexType) and not isinstance(
                    self.data[row_ndx][col_ndx], _TableIndex
                ):
                    self.data[row_ndx][col_ndx] = _TableIndex(
                        self.data[row_ndx][col_ndx]
                    )
                    self.data[row_ndx][col_ndx].set_table(
                        c_types[fk_col].params["table"]
                    )

            # Wrap the Primary Key
            if self._pk_col is not None:
                col_ndx = self.columns.index(self._pk_col)
                self.data[row_ndx][col_ndx] = _TableKey(self.data[row_ndx][col_ndx])
                self.data[row_ndx][col_ndx].set_table(self, self._pk_col)

        if only_last:
            update_row(len(self.data) - 1)
        else:
            for row_ndx in range(len(self.data)):
                update_row(row_ndx)

    def add_column(self, name, data, optional=False):
        """Adds a column of data to the table.

        Args:
            name: (str) - the unique name of the column
            data: (list | np.array) - a column of homogeneous data
            optional: (bool) - if null-like values are permitted
        """
        assert isinstance(name, str) and name not in self.columns
        is_np = util.is_numpy_array(data)
        assert isinstance(data, list) or is_np
        assert isinstance(optional, bool)
        is_first_col = len(self.columns) == 0
        assert is_first_col or len(data) == len(
            self.data
        ), f"Expected length {len(self.data)}, found {len(data)}"

        # Add the new data
        for ndx in range(max(len(data), len(self.data))):
            if is_first_col:
                self.data.append([])
            if is_np:
                self.data[ndx].append(data[ndx])
            else:
                self.data[ndx].append(data[ndx])
        # add the column
        self.columns.append(name)

        try:
            self.cast(name, _dtypes.UnknownType(), optional=optional)
        except TypeError as err:
            # Undo the changes
            if is_first_col:
                self.data = []
                self.columns = []
            else:
                for ndx in range(len(self.data)):
                    self.data[ndx] = self.data[ndx][:-1]
                self.columns = self.columns[:-1]
            raise err

    def get_column(self, name, convert_to=None):
        """Retrieves a column from the table and optionally converts it to a NumPy object.

        Args:
            name: (str) - the name of the column
            convert_to: (str, optional)
                - "numpy": will convert the underlying data to numpy object
        """
        assert name in self.columns
        assert convert_to is None or convert_to == "numpy"
        if convert_to == "numpy":
            np = util.get_module(
                "numpy", required="Converting to NumPy requires installing NumPy"
            )
        col = []
        col_ndx = self.columns.index(name)
        for row in self.data:
            item = row[col_ndx]
            if convert_to is not None and isinstance(item, WBValue):
                item = item.to_data_array()
            col.append(item)
        if convert_to == "numpy":
            col = np.array(col)
        return col

    def get_index(self):
        """Returns an array of row indexes for use in other tables to create links."""
        ndxs = []
        for ndx in range(len(self.data)):
            index = _TableIndex(ndx)
            index.set_table(self)
            ndxs.append(index)
        return ndxs

    def get_dataframe(self):
        """Returns a `pandas.DataFrame` of the table."""
        pd = util.get_module(
            "pandas",
            required="Converting to pandas.DataFrame requires installing pandas",
        )
        return pd.DataFrame.from_records(self.data, columns=self.columns)

    def index_ref(self, index):
        """Gets a reference of the index of a row in the table."""
        assert index < len(self.data)
        _index = _TableIndex(index)
        _index.set_table(self)
        return _index

    def add_computed_columns(self, fn):
        """Adds one or more computed columns based on existing data.

        Args:
            fn: A function which accepts one or two parameters, ndx (int) and row (dict),
                which is expected to return a dict representing new columns for that row, keyed
                by the new column names.

                `ndx` is an integer representing the index of the row. Only included if `include_ndx`
                      is set to `True`.

                `row` is a dictionary keyed by existing columns
        """
        new_columns = {}
        for ndx, row in self.iterrows():
            row_dict = {self.columns[i]: row[i] for i in range(len(self.columns))}
            new_row_dict = fn(ndx, row_dict)
            assert isinstance(new_row_dict, dict)
            for key in new_row_dict:
                new_columns[key] = new_columns.get(key, [])
                new_columns[key].append(new_row_dict[key])
        for new_col_name in new_columns:
            self.add_column(new_col_name, new_columns[new_col_name])


class _PartitionTablePartEntry:
    """Helper class for PartitionTable to track its parts."""

    def __init__(self, entry, source_artifact):
        self.entry = entry
        self.source_artifact = source_artifact
        self._part = None

    def get_part(self):
        if self._part is None:
            self._part = self.source_artifact.get(self.entry.path)
        return self._part

    def free(self):
        self._part = None


class PartitionedTable(Media):
    """A table which is composed of multiple sub-tables.

    Currently, PartitionedTable is designed to point to a directory within an artifact.
    """

    _log_type = "partitioned-table"

    def __init__(self, parts_path):
        """Initialize a PartitionedTable.

        Args:
            parts_path (str): path to a directory of tables in the artifact.
        """
        super().__init__()
        self.parts_path = parts_path
        self._loaded_part_entries = {}

    def to_json(self, artifact_or_run):
        json_obj = {
            "_type": PartitionedTable._log_type,
        }
        if isinstance(artifact_or_run, wandb.wandb_sdk.wandb_run.Run):
            artifact_entry_url = self._get_artifact_entry_ref_url()
            if artifact_entry_url is None:
                raise ValueError(
                    "PartitionedTables must first be added to an Artifact before logging to a Run"
                )
            json_obj["artifact_path"] = artifact_entry_url
        else:
            json_obj["parts_path"] = self.parts_path
        return json_obj

    @classmethod
    def from_json(cls, json_obj, source_artifact):
        instance = cls(json_obj["parts_path"])
        entries = source_artifact.manifest.get_entries_in_directory(
            json_obj["parts_path"]
        )
        for entry in entries:
            instance._add_part_entry(entry, source_artifact)
        return instance

    def iterrows(self):
        """Iterate over rows as (ndx, row).

        Yields:
        ------
        index : int
            The index of the row.
        row : List[any]
            The data of the row.
        """
        columns = None
        ndx = 0
        for entry_path in self._loaded_part_entries:
            part = self._loaded_part_entries[entry_path].get_part()
            if columns is None:
                columns = part.columns
            elif columns != part.columns:
                raise ValueError(
                    "Table parts have non-matching columns. {} != {}".format(
                        columns, part.columns
                    )
                )
            for _, row in part.iterrows():
                yield ndx, row
                ndx += 1

            self._loaded_part_entries[entry_path].free()

    def _add_part_entry(self, entry, source_artifact):
        self._loaded_part_entries[entry.path] = _PartitionTablePartEntry(
            entry, source_artifact
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.parts_path == other.parts_path

    def bind_to_run(self, *args, **kwargs):
        raise ValueError("PartitionedTables cannot be bound to runs")


class JoinedTable(Media):
    """Join two tables for visualization in the Artifact UI.

    Args:
        table1 (str, wandb.Table, ArtifactManifestEntry):
            the path to a wandb.Table in an artifact, the table object, or ArtifactManifestEntry
        table2 (str, wandb.Table):
            the path to a wandb.Table in an artifact, the table object, or ArtifactManifestEntry
        join_key (str, [str, str]):
            key or keys to perform the join
    """

    _log_type = "joined-table"

    def __init__(self, table1, table2, join_key):
        super().__init__()

        if not isinstance(join_key, str) and (
            not isinstance(join_key, list) or len(join_key) != 2
        ):
            raise ValueError(
                "JoinedTable join_key should be a string or a list of two strings"
            )

        if not self._validate_table_input(table1):
            raise ValueError(
                "JoinedTable table1 should be an artifact path to a table or wandb.Table object"
            )

        if not self._validate_table_input(table2):
            raise ValueError(
                "JoinedTable table2 should be an artifact path to a table or wandb.Table object"
            )

        self._table1 = table1
        self._table2 = table2
        self._join_key = join_key

    @classmethod
    def from_json(cls, json_obj, source_artifact):
        t1 = source_artifact.get(json_obj["table1"])
        if t1 is None:
            t1 = json_obj["table1"]

        t2 = source_artifact.get(json_obj["table2"])
        if t2 is None:
            t2 = json_obj["table2"]

        return cls(
            t1,
            t2,
            json_obj["join_key"],
        )

    @staticmethod
    def _validate_table_input(table):
        """Helper method to validate that the table input is one of the 3 supported types."""
        return (
            (isinstance(table, str) and table.endswith(".table.json"))
            or isinstance(table, Table)
            or isinstance(table, PartitionedTable)
            or (hasattr(table, "ref_url") and table.ref_url().endswith(".table.json"))
        )

    def _ensure_table_in_artifact(self, table, artifact, table_ndx):
        """Helper method to add the table to the incoming artifact. Returns the path."""
        if isinstance(table, Table) or isinstance(table, PartitionedTable):
            table_name = f"t{table_ndx}_{str(id(self))}"
            if (
                table._artifact_source is not None
                and table._artifact_source.name is not None
            ):
                table_name = os.path.basename(table._artifact_source.name)
            entry = artifact.add(table, table_name)
            table = entry.path
        # Check if this is an ArtifactManifestEntry
        elif hasattr(table, "ref_url"):
            # Give the new object a unique, yet deterministic name
            name = binascii.hexlify(base64.standard_b64decode(table.digest)).decode(
                "ascii"
            )[:20]
            entry = artifact.add_reference(
                table.ref_url(), "{}.{}.json".format(name, table.name.split(".")[-2])
            )[0]
            table = entry.path

        err_str = "JoinedTable table:{} not found in artifact. Add a table to the artifact using Artifact#add(<table>, {}) before adding this JoinedTable"
        if table not in artifact._manifest.entries:
            raise ValueError(err_str.format(table, table))

        return table

    def to_json(self, artifact_or_run):
        json_obj = {
            "_type": JoinedTable._log_type,
        }
        if isinstance(artifact_or_run, wandb.wandb_sdk.wandb_run.Run):
            artifact_entry_url = self._get_artifact_entry_ref_url()
            if artifact_entry_url is None:
                raise ValueError(
                    "JoinedTables must first be added to an Artifact before logging to a Run"
                )
            json_obj["artifact_path"] = artifact_entry_url
        else:
            table1 = self._ensure_table_in_artifact(self._table1, artifact_or_run, 1)
            table2 = self._ensure_table_in_artifact(self._table2, artifact_or_run, 2)
            json_obj.update(
                {
                    "table1": table1,
                    "table2": table2,
                    "join_key": self._join_key,
                }
            )
        return json_obj

    def __ne__(self, other):
        return not self.__eq__(other)

    def _eq_debug(self, other, should_assert=False):
        eq = isinstance(other, JoinedTable)
        assert not should_assert or eq, "Found type {}, expected {}".format(
            other.__class__, JoinedTable
        )
        eq = eq and self._join_key == other._join_key
        assert not should_assert or eq, "Found {} join key, expected {}".format(
            other._join_key, self._join_key
        )
        eq = eq and self._table1._eq_debug(other._table1, should_assert)
        eq = eq and self._table2._eq_debug(other._table2, should_assert)
        return eq

    def __eq__(self, other):
        return self._eq_debug(other, False)

    def bind_to_run(self, *args, **kwargs):
        raise ValueError("JoinedTables cannot be bound to runs")


class _TableType(_dtypes.Type):
    name = "table"
    legacy_names = ["wandb.Table"]
    types = [Table]

    def __init__(self, column_types=None):
        if column_types is None:
            column_types = _dtypes.UnknownType()
        if isinstance(column_types, dict):
            column_types = _dtypes.TypedDictType(column_types)
        elif not (
            isinstance(column_types, _dtypes.TypedDictType)
            or isinstance(column_types, _dtypes.UnknownType)
        ):
            raise TypeError("column_types must be a dict or TypedDictType")

        self.params.update({"column_types": column_types})

    def assign_type(self, wb_type=None):
        if isinstance(wb_type, _TableType):
            column_types = self.params["column_types"].assign_type(
                wb_type.params["column_types"]
            )
            if not isinstance(column_types, _dtypes.InvalidType):
                return _TableType(column_types)

        return _dtypes.InvalidType()

    @classmethod
    def from_obj(cls, py_obj):
        if not isinstance(py_obj, Table):
            raise TypeError("py_obj must be a wandb.Table")
        else:
            return cls(py_obj._column_types)


class _JoinedTableType(_dtypes.Type):
    name = "joined-table"
    types = [JoinedTable]


class _PartitionedTableType(_dtypes.Type):
    name = "partitioned-table"
    types = [PartitionedTable]


_dtypes.TypeRegistry.add(_TableType)
_dtypes.TypeRegistry.add(_JoinedTableType)
_dtypes.TypeRegistry.add(_PartitionedTableType)
_dtypes.TypeRegistry.add(_ForeignKeyType)
_dtypes.TypeRegistry.add(_PrimaryKeyType)
_dtypes.TypeRegistry.add(_ForeignIndexType)



import functools
import logging
import os
from io import BytesIO
from typing import TYPE_CHECKING, Any, Literal, Optional, Sequence, Type, Union

import wandb
from wandb import util
from wandb.sdk.lib import filesystem, printer, printer_asyncio, runid

from . import _dtypes
from ._private import MEDIA_TMP
from .base_types.media import BatchableMedia

if TYPE_CHECKING:  # pragma: no cover
    from typing import TextIO

    import numpy as np

    from wandb.sdk.artifacts.artifact import Artifact

    from ..wandb_run import Run as LocalRun


# This helper function is a workaround for the issue discussed here:
# https://github.com/wandb/wandb/issues/3472
#
# Essentially, the issue is that moviepy's write_gif function fails to close
# the open write / file descriptor returned from `imageio.save`. The following
# function is a simplified copy of the function in the moviepy source code.
# See https://github.com/Zulko/moviepy/blob/7e3e8bb1b739eb6d1c0784b0cb2594b587b93b39/moviepy/video/io/gif_writers.py#L428
#
# Except, we close the writer!
def write_gif_with_image_io(
    clip: Any, filename: str, fps: Optional[int] = None
) -> None:
    imageio = util.get_module(
        "imageio",
        required='wandb.Video requires imageio when passing raw data. Install with "pip install wandb[media]"',
    )

    writer = imageio.save(filename, fps=clip.fps, quantizer=0, palettesize=256, loop=0)

    for frame in clip.iter_frames(fps=fps, dtype="uint8"):
        writer.append_data(frame)

    writer.close()


class Video(BatchableMedia):
    """A class for logging videos to W&B."""

    _log_type = "video-file"
    EXTS = ("gif", "mp4", "webm", "ogg")
    _width: Optional[int]
    _height: Optional[int]

    def __init__(
        self,
        data_or_path: Union["np.ndarray", str, "TextIO", "BytesIO"],
        caption: Optional[str] = None,
        fps: Optional[int] = None,
        format: Optional[Literal["gif", "mp4", "webm", "ogg"]] = None,
    ):
        """Initialize a W&B Video object.

        Args:
            data_or_path:
                Video can be initialized with a path to a file or an io object.
                Video can be initialized with a numpy tensor.
                The numpy tensor must be either 4 dimensional or 5 dimensional.
                The dimensions should be (number of frames, channel, height, width) or
                (batch, number of frames, channel, height, width)
                The format parameter must be specified with the format argument
                when initializing with a numpy array
                or io object.
            caption: Caption associated with the video for display.
            fps:
                The frame rate to use when encoding raw video frames.
                Default value is 4.
                This parameter has no effect when data_or_path is a string, or bytes.
            format:
                Format of video, necessary if initializing with a numpy array
                or io object. This parameter will be used to determine the format
                to use when encoding the video data. Accepted values are "gif",
                "mp4", "webm", or "ogg".

        Examples:
            ### Log a numpy array as a video
            ```python
            import numpy as np
            import wandb

            with wandb.init() as run:
                # axes are (number of frames, channel, height, width)
                frames = np.random.randint(
                    low=0, high=256, size=(10, 3, 100, 100), dtype=np.uint8
                )
                run.log({"video": wandb.Video(frames, format="mp4", fps=4)})
            ```
        """
        super().__init__(caption=caption)

        if format is None:
            wandb.termwarn(
                "`format` argument was not provided, defaulting to `gif`. "
                "This parameter will be required in v0.20.0, "
                "please specify the format explicitly."
            )
        self._format = format or "gif"
        self._width = None
        self._height = None
        self._channels = None
        if self._format not in Video.EXTS:
            raise ValueError(
                "wandb.Video accepts {} formats".format(", ".join(Video.EXTS))
            )

        if isinstance(data_or_path, (BytesIO, str)) and fps:
            msg = (
                "`fps` argument does not affect the frame rate of the video "
                "when providing a file path or raw bytes."
            )
            wandb.termwarn(msg)

        if isinstance(data_or_path, BytesIO):
            filename = os.path.join(
                MEDIA_TMP.name, runid.generate_id() + "." + self._format
            )
            with open(filename, "wb") as f:
                f.write(data_or_path.read())
            self._set_file(filename, is_tmp=True)
        elif isinstance(data_or_path, str):
            _, ext = os.path.splitext(data_or_path)
            ext = ext[1:].lower()
            if ext not in Video.EXTS:
                raise ValueError(
                    "wandb.Video accepts {} formats".format(", ".join(Video.EXTS))
                )
            self._set_file(data_or_path, is_tmp=False)
            # ffprobe -v error -select_streams v:0 -show_entries stream=width,height -of csv=p=0 data_or_path
        else:
            if hasattr(data_or_path, "numpy"):  # TF data eager tensors
                self.data = data_or_path.numpy()
            elif util.is_numpy_array(data_or_path):
                self.data = data_or_path
            else:
                raise ValueError(
                    "wandb.Video accepts a file path or numpy like data as input"
                )
            fps = fps or 4
            printer_asyncio.run_async_with_spinner(
                printer.new_printer(),
                "Encoding video...",
                functools.partial(self.encode, fps=fps),
            )

    def encode(self, fps: int = 4) -> None:
        # import ImageSequenceClip from the appropriate MoviePy module
        mpy = util.get_module(
            "moviepy.video.io.ImageSequenceClip",
            required='wandb.Video requires moviepy when passing raw data. Install with "pip install wandb[media]"',
        )

        tensor = self._prepare_video(self.data)
        _, self._height, self._width, self._channels = tensor.shape  # type: ignore

        # encode sequence of images into gif string
        clip = mpy.ImageSequenceClip(list(tensor), fps=fps)

        filename = os.path.join(
            MEDIA_TMP.name, runid.generate_id() + "." + self._format
        )

        if self._format == "gif":
            write_gif_with_image_io(clip, filename)
        else:
            clip.write_videofile(filename, logger=None)

        self._set_file(filename, is_tmp=True)

    @classmethod
    def get_media_subdir(cls: Type["Video"]) -> str:
        return os.path.join("media", "videos")

    def to_json(self, run_or_artifact: Union["LocalRun", "Artifact"]) -> dict:
        json_dict = super().to_json(run_or_artifact)
        json_dict["_type"] = self._log_type

        if self._width is not None:
            json_dict["width"] = self._width
        if self._height is not None:
            json_dict["height"] = self._height

        return json_dict

    def _prepare_video(self, video: "np.ndarray") -> "np.ndarray":
        """This logic was mostly taken from tensorboardX."""
        np = util.get_module(
            "numpy",
            required='wandb.Video requires numpy when passing raw data. To get it, run "pip install numpy".',
        )
        if video.ndim < 4:
            raise ValueError(
                "Video must be at least 4 dimensions: time, channels, height, width"
            )
        if video.ndim == 4:
            video = video.reshape(1, *video.shape)
        b, t, c, h, w = video.shape

        if video.dtype != np.uint8:
            logging.warning("Converting video data to uint8")
            video = video.astype(np.uint8)

        def is_power2(num: int) -> bool:
            return num != 0 and ((num & (num - 1)) == 0)

        # pad to nearest power of 2, all at once
        if not is_power2(video.shape[0]):
            len_addition = int(2 ** video.shape[0].bit_length() - video.shape[0])
            video = np.concatenate(
                (video, np.zeros(shape=(len_addition, t, c, h, w))), axis=0
            )

        n_rows = 2 ** ((b.bit_length() - 1) // 2)
        n_cols = video.shape[0] // n_rows

        video = video.reshape(n_rows, n_cols, t, c, h, w)
        video = np.transpose(video, axes=(2, 0, 4, 1, 5, 3))
        video = video.reshape(t, n_rows * h, n_cols * w, c)
        return video

    @classmethod
    def seq_to_json(
        cls: Type["Video"],
        seq: Sequence["BatchableMedia"],
        run: "LocalRun",
        key: str,
        step: Union[int, str],
    ) -> dict:
        base_path = os.path.join(run.dir, cls.get_media_subdir())
        filesystem.mkdir_exists_ok(base_path)

        meta = {
            "_type": "videos",
            "count": len(seq),
            "videos": [v.to_json(run) for v in seq],
            "captions": Video.captions(seq),
        }
        return meta


class _VideoFileType(_dtypes.Type):
    name = "video-file"
    types = [Video]


_dtypes.TypeRegistry.add(_VideoFileType)



from __future__ import annotations

import logging
import secrets
import string
import threading

from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_server_pb2 as spb

from .mailbox_handle import MailboxHandle
from .response_handle import MailboxResponseHandle

_logger = logging.getLogger(__name__)


class MailboxClosedError(Exception):
    """The mailbox has been closed and cannot be used."""


class Mailbox:
    """Matches service responses to requests.

    The mailbox can set an address on a server request and create a handle for
    waiting for a response to that record. Responses are delivered by calling
    `deliver()`. The `close()` method abandons all handles in case the
    service process becomes unreachable.
    """

    def __init__(self) -> None:
        self._handles: dict[str, MailboxResponseHandle] = {}
        self._handles_lock = threading.Lock()
        self._closed = False

    def require_response(
        self,
        request: spb.ServerRequest | pb.Record,
    ) -> MailboxHandle[spb.ServerResponse]:
        """Set a response address on a request.

        Args:
            request: The request on which to set a request ID or mailbox slot.
                This is mutated. An address must not already be set.

        Returns:
            A handle for waiting for the response to the request.

        Raises:
            MailboxClosedError: If the mailbox has been closed, in which case
                no new responses are expected to be delivered and new handles
                cannot be created.
        """
        if isinstance(request, spb.ServerRequest):
            if address := request.request_id:
                raise ValueError(f"Request already has an address ({address})")

            address = self._new_address()
            request.request_id = address
        else:
            if address := request.control.mailbox_slot:
                raise ValueError(f"Request already has an address ({address})")

            address = self._new_address()
            request.control.mailbox_slot = address

        with self._handles_lock:
            if self._closed:
                raise MailboxClosedError()

            handle = MailboxResponseHandle(address)
            self._handles[address] = handle

        return handle

    def _new_address(self) -> str:
        """Returns an unused address for a request.

        Assumes `_handles_lock` is held.
        """

        def generate():
            return "".join(
                secrets.choice(string.ascii_lowercase + string.digits)
                for i in range(12)
            )

        address = generate()

        # Being extra cautious. This loop will almost never be entered.
        while address in self._handles:
            address = generate()

        return address

    def deliver(self, response: spb.ServerResponse) -> None:
        """Deliver a response from the service.

        If the response address is invalid, this does nothing.
        It is a no-op if the mailbox has been closed.
        """
        address = response.request_id
        if not address:
            kind: str | None = response.WhichOneof("server_response_type")
            if kind == "result_communicate":
                result_type = response.result_communicate.WhichOneof("result_type")
                kind = f"result_communicate.{result_type}"

            _logger.error(f"Received response with no mailbox slot: {kind}")
            return

        with self._handles_lock:
            # NOTE: If the mailbox is closed, this returns None because
            # we clear the dict.
            handle = self._handles.pop(address, None)

        # It is not an error if there is no handle for the address:
        # handles can be abandoned if the result is no longer needed.
        if handle:
            handle.deliver(response)

    def close(self) -> None:
        """Indicate no further responses will be delivered.

        Abandons all handles.
        """
        with self._handles_lock:
            self._closed = True

            _logger.info(
                f"Closing mailbox, abandoning {len(self._handles)} handles.",
            )

            for handle in self._handles.values():
                handle.abandon()
            self._handles.clear()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import List, Optional

from pydantic import Field

from wandb._pydantic import GQLBase, GQLId


class CreateArtifactCollectionTagAssignments(GQLBase):
    create_artifact_collection_tag_assignments: Optional[
        CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignments
    ] = Field(alias="createArtifactCollectionTagAssignments")


class CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignments(
    GQLBase
):
    tags: List[
        CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignmentsTags
    ]


class CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignmentsTags(
    GQLBase
):
    id: GQLId
    name: str
    tag_category_name: str = Field(alias="tagCategoryName")


CreateArtifactCollectionTagAssignments.model_rebuild()
CreateArtifactCollectionTagAssignmentsCreateArtifactCollectionTagAssignments.model_rebuild()



from __future__ import annotations

from textwrap import dedent

from wandb_graphql.language.printer import print_ast

from wandb.apis.public.utils import gql_compat
from wandb.sdk.internal.internal_api import Api as InternalApi

OMITTABLE_ARTIFACT_FIELDS = frozenset(
    {
        "ttlDurationSeconds",
        "ttlIsInherited",
        "aliases",
        "tags",
        "historyStep",
    }
)


def omit_artifact_fields(api: InternalApi) -> set[str]:
    """Return names of Artifact fields to remove from GraphQL requests (for server compatibility)."""
    allowed_fields = set(api.server_artifact_introspection())
    return set(OMITTABLE_ARTIFACT_FIELDS - allowed_fields)


def _gql_artifact_fragment(include_aliases: bool = True) -> str:
    """Return a GraphQL query fragment with all parseable Artifact attributes."""
    omit_fields = omit_artifact_fields(api=InternalApi())

    # Respect the `include_aliases` flag
    if not include_aliases:
        omit_fields.add("aliases")

    artifact_fragment_str = dedent(
        """\
        fragment ArtifactFragment on Artifact {
            id
            artifactSequence {
                project {
                    entityName
                    name
                }
                name
            }
            versionIndex
            artifactType {
                name
            }
            description
            metadata
            ttlDurationSeconds
            ttlIsInherited
            aliases {
                artifactCollection {
                    project {
                        entityName
                        name
                    }
                    name
                }
                alias
            }
            tags {
                name
            }
            historyStep
            state
            currentManifest {
                file {
                    directUrl
                }
            }
            commitHash
            fileCount
            createdAt
            updatedAt
        }"""
    )
    compat_doc = gql_compat(artifact_fragment_str, omit_fields=omit_fields)
    return print_ast(compat_doc)


def _gql_registry_fragment() -> str:
    return """
        fragment RegistryFragment on Project {
           id
            allowAllArtifactTypesInRegistry
            artifactTypes(includeAll: true) {
                edges {
                    node {
                        name
                    }
                }
            }
            name
            description
            createdAt
            updatedAt
            access
        }
    """



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLBase


class DeleteAliases(GQLBase):
    delete_aliases: Optional[DeleteAliasesDeleteAliases] = Field(alias="deleteAliases")


class DeleteAliasesDeleteAliases(GQLBase):
    success: bool


DeleteAliases.model_rebuild()



"""Multi storage handler."""

from __future__ import annotations

from typing import TYPE_CHECKING, Sequence
from urllib.parse import urlparse

from wandb.sdk.artifacts.storage_handler import StorageHandler
from wandb.sdk.lib.paths import FilePathStr, URIStr

if TYPE_CHECKING:
    from wandb.sdk.artifacts.artifact import Artifact
    from wandb.sdk.artifacts.artifact_manifest_entry import ArtifactManifestEntry


class MultiHandler(StorageHandler):
    _handlers: list[StorageHandler]

    def __init__(
        self,
        handlers: list[StorageHandler] | None = None,
        default_handler: StorageHandler | None = None,
    ) -> None:
        self._handlers = handlers or []
        self._default_handler = default_handler

    def _get_handler(self, url: FilePathStr | URIStr) -> StorageHandler:
        parsed_url = urlparse(url)
        for handler in self._handlers:
            if handler.can_handle(parsed_url):
                return handler
        if self._default_handler is not None:
            return self._default_handler
        raise ValueError('No storage handler registered for url "{}"'.format(str(url)))

    def load_path(
        self,
        manifest_entry: ArtifactManifestEntry,
        local: bool = False,
    ) -> URIStr | FilePathStr:
        assert manifest_entry.ref is not None
        handler = self._get_handler(manifest_entry.ref)
        return handler.load_path(manifest_entry, local=local)

    def store_path(
        self,
        artifact: Artifact,
        path: URIStr | FilePathStr,
        name: str | None = None,
        checksum: bool = True,
        max_objects: int | None = None,
    ) -> Sequence[ArtifactManifestEntry]:
        handler = self._get_handler(path)
        return handler.store_path(
            artifact, path, name=name, checksum=checksum, max_objects=max_objects
        )



"""module server."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from wandb import util
from wandb.apis import InternalApi

if TYPE_CHECKING:
    from wandb.sdk.wandb_settings import Settings


class Server:
    def __init__(
        self,
        settings: Settings,
    ) -> None:
        self._api = InternalApi(default_settings=settings)
        self._viewer: dict[str, Any] = {}
        self._flags: dict[str, Any] = {}
        self._settings: Settings = settings

    def query_with_timeout(self, timeout: int | float = 5) -> None:
        if self._settings.x_disable_viewer:
            return

        async_viewer = util.async_call(self._api.viewer_server_info, timeout=timeout)
        try:
            viewer_tuple, viewer_thread = async_viewer()
        except Exception:
            return
        if viewer_thread.is_alive():
            # this is likely a DNS hang
            return
        # TODO(jhr): should we kill the thread?
        self._viewer, self._serverinfo = viewer_tuple
        self._flags = json.loads(self._viewer.get("flags", "{}"))

    @property
    def viewer(self) -> dict[str, Any]:
        """Returns information about the currently authenticated user.

        If the API key is valid, the following is returned:
        - id
        - entity
        - username
        - flags
        - teams

        If the API key is not valid or the server is not reachable,
        an empty dict is returned.
        """
        if not self._viewer and not self._settings._offline:
            self.query_with_timeout()

        return self._viewer



# ruff: noqa: UP007  # Avoid using `X | Y` for union fields, as this can cause issues with pydantic < 2.6

from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import Field
from typing_extensions import Annotated

from wandb._pydantic import GQLBase, GQLId

from ._generated import TriggerFields
from .actions import InputAction, SavedAction
from .events import InputEvent, SavedEvent
from .scopes import AutomationScope


# ------------------------------------------------------------------------------
# Saved types: for parsing response data from saved automations
class Automation(TriggerFields):
    """A local instance of a saved W&B automation."""

    id: GQLId

    created_at: Annotated[datetime, Field(repr=False, frozen=True, alias="createdAt")]
    """The date and time when this automation was created."""

    updated_at: Annotated[
        Optional[datetime], Field(repr=False, frozen=True, alias="updatedAt")
    ] = None
    """The date and time when this automation was last updated, if applicable."""

    name: str
    """The name of this automation."""

    description: Optional[str]
    """An optional description of this automation."""

    enabled: bool
    """Whether this automation is enabled.  Only enabled automations will trigger."""

    event: SavedEvent
    """The event that will trigger this automation."""

    scope: AutomationScope
    """The scope in which the triggering event must occur."""

    action: SavedAction
    """The action that will execute when this automation is triggered."""


class NewAutomation(GQLBase, extra="forbid", validate_default=False):
    """A new automation to be created."""

    name: Optional[str] = None
    """The name of this automation."""

    description: Optional[str] = None
    """An optional description of this automation."""

    enabled: Optional[bool] = None
    """Whether this automation is enabled.  Only enabled automations will trigger."""

    event: Optional[InputEvent] = None
    """The event that will trigger this automation."""

    # Ensure that the event and its scope are always consistent, if the event is set.
    @property
    def scope(self) -> Optional[AutomationScope]:
        """The scope in which the triggering event must occur."""
        return self.event.scope if self.event else None

    @scope.setter
    def scope(self, value: AutomationScope) -> None:
        if self.event is None:
            raise ValueError("Cannot set `scope` for an automation with no `event`")
        self.event.scope = value

    action: Optional[InputAction] = None
    """The action that will execute when this automation is triggered."""


__all__ = [
    "Automation",
    "NewAutomation",
]
